# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from blueapps.account.sites.default import ConfFixture as default_fixture
from blueapps.settings import blueapps_settings


class _ConfFixture:
    ACCOUNT_CONF_PREFIX = "ACCOUNT"

    def __init__(self, fixture_module):
        # store the module
        self._fixture = import_string(fixture_module)

    def __getattr__(self, name):
        # settings fixture, 优先返回开发者配置
        account_conf_key = f"{self.ACCOUNT_CONF_PREFIX}_{name}"
        if hasattr(blueapps_settings, account_conf_key):
            return getattr(blueapps_settings, account_conf_key)

        # site fixture, 对应环境配置
        if hasattr(self._fixture, name):
            return getattr(self._fixture, name)

        # default fixture, 默认配置
        if hasattr(default_fixture, name):
            setting = getattr(default_fixture, name)
            if setting is None:
                raise ImproperlyConfigured(
                    "Requested %s, but ConfFixture are not configured. "
                    "You must set options in ConfFixture in right site.conf.py" % name
                )
            return setting

        raise KeyError("%s not exist" % name)


MOD = "blueapps.account.sites.{VER}.conf.ConfFixture".format(VER=settings.RUN_VER)
ConfFixture = _ConfFixture(MOD)

AUTH_USER_MODEL = "account.User"
# apigw cache key with 1 year
APIGW_CACHE_KEY = "public_key_cache"
APIGW_CACHE_EXPIRES = 1 * 365 * 86400

######################
# 二次验证配置默认参数 #
######################

# 短信验证有效时间
SECOND_VERIFY_CONF = {
    "VALID_MINUTES": 5,
    "RETRY_MINUTES": 3,
    "SMS_FORMAT": _("您正在蓝鲸应用上执行敏感操作，验证码：{}"),
    "CODE_NAME": "bk_verify_code",
}



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
from json.decoder import JSONDecodeError

from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import exception_handler

from blueapps.core.exceptions import BlueException
from blueapps.utils.logger import logger


def custom_exception_handler(exc, context):
    """
    分类：
        APIException及子类异常
        app自定义异常和未处理异常
    """
    response = exception_handler(exc, context)
    if response:
        return Response(
            response.data["detail"] if "detail" in response.data else response.data, status=response.status_code,
        )

    exc_message = str(exc)
    exc_data = None
    if hasattr(exc, "data") and exc.data:
        try:
            exc_data = json.loads(exc.data)
        except JSONDecodeError:
            exc_data = exc.data
        except TypeError:
            # 其它内容不能被json解析 忽略
            pass

    if hasattr(exc, "message") and exc.message:
        try:
            exc_message = json.loads(str(exc.message))
        except JSONDecodeError:
            exc_message = str(exc.message)

    code = status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    if isinstance(exc, BlueException):
        code = exc.code
        status_code = exc.STATUS_CODE

    data = {
        "code":  code,
        "message": exc_message,
        "data": exc_data,
    }
    # 使用json方便提取
    logger.exception(json.dumps({"code": data["code"], "message": data["message"], "data": data["data"]}))
    return Response(data, status=status_code)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import models
from django.utils.translation import ugettext_lazy as _


class DjCrontabSchedule(models.Model):
    minute = models.CharField(max_length=64, default="*")
    hour = models.CharField(max_length=64, default="*")
    day_of_week = models.CharField(max_length=64, default="*")
    day_of_month = models.CharField(max_length=64, default="*")
    month_of_year = models.CharField(max_length=64, default="*")

    class Meta:
        db_table = "djcelery_crontabschedule"


class DjPeriodicTasks(models.Model):
    ident = models.SmallIntegerField(default=1, primary_key=True, unique=True)
    last_update = models.DateTimeField(null=False)

    class Meta:
        db_table = "djcelery_periodictasks"


class DjIntervalSchedule(models.Model):
    every = models.IntegerField(_("every"), null=False)
    period = models.CharField(_("period"), max_length=24)

    class Meta:
        db_table = "djcelery_intervalschedule"
        verbose_name = _("interval")
        verbose_name_plural = _("intervals")
        ordering = ["period", "every"]


class DjPeriodicTask(models.Model):
    name = models.CharField(_("name"), max_length=200, unique=True, help_text=_("Useful description"), )
    task = models.CharField(_("task name"), max_length=200)
    interval = models.ForeignKey(
        DjIntervalSchedule,
        null=True,
        blank=True,
        verbose_name=_("interval"),
        on_delete=models.CASCADE,
    )
    crontab = models.ForeignKey(
        DjCrontabSchedule,
        null=True,
        blank=True,
        verbose_name=_("crontab"),
        on_delete=models.CASCADE,
        help_text=_("Use one of interval/crontab"),
    )
    args = models.TextField(
        _("Arguments"),
        blank=True,
        default="[]",
        help_text=_("JSON encoded positional arguments"),
    )
    kwargs = models.TextField(
        _("Keyword arguments"),
        blank=True,
        default="{}",
        help_text=_("JSON encoded keyword arguments"),
    )
    queue = models.CharField(
        _("queue"),
        max_length=200,
        blank=True,
        null=True,
        default=None,
        help_text=_("Queue defined in CELERY_QUEUES"),
    )
    exchange = models.CharField(_("exchange"), max_length=200, blank=True, null=True, default=None, )
    routing_key = models.CharField(_("routing key"), max_length=200, blank=True, null=True, default=None, )
    expires = models.DateTimeField(_("expires"), blank=True, null=True)
    enabled = models.BooleanField(_("enabled"), default=True)
    last_run_at = models.DateTimeField(
        auto_now=False,
        auto_now_add=False,
        editable=False,
        blank=True,
        null=True,
    )
    total_run_count = models.PositiveIntegerField(default=0, editable=False)
    date_changed = models.DateTimeField(auto_now=True)
    description = models.TextField(_("description"), blank=True)

    no_changes = False

    class Meta:
        db_table = "djcelery_periodictask"
        verbose_name = _("periodic task")
        verbose_name_plural = _("periodic tasks")


class DjWorkerState(models.Model):
    hostname = models.CharField(_("hostname"), max_length=255, unique=True)
    last_heartbeat = models.DateTimeField(_("last heartbeat"), null=True, db_index=True)

    class Meta:
        """Model meta-data."""

        verbose_name = _("worker")
        verbose_name_plural = _("workers")
        get_latest_by = "last_heartbeat"
        ordering = ["-last_heartbeat"]


class DjTaskState(models.Model):
    state = models.CharField(_("state"), max_length=64)
    task_id = models.CharField(_("UUID"), max_length=36, unique=True)
    name = models.CharField(_("name"), max_length=200, null=True, db_index=True)
    tstamp = models.DateTimeField(_("event received at"), db_index=True)
    args = models.TextField(_("Arguments"), null=True)
    kwargs = models.TextField(_("Keyword arguments"), null=True)
    eta = models.DateTimeField(_("ETA"), null=True)
    expires = models.DateTimeField(_("expires"), null=True)
    result = models.TextField(_("result"), null=True)
    traceback = models.TextField(_("traceback"), null=True)
    runtime = models.FloatField(
        _("execution time"),
        null=True,
        help_text=_("in seconds if task succeeded"),
    )
    retries = models.IntegerField(_("number of retries"), default=0)
    worker = models.ForeignKey(
        DjWorkerState,
        null=True,
        verbose_name=_("worker"),
        on_delete=models.CASCADE,
    )
    hidden = models.BooleanField(editable=False, default=False, db_index=True)

    class Meta:
        """Model meta-data."""

        verbose_name = _("task")
        verbose_name_plural = _("tasks")
        get_latest_by = "tstamp"
        ordering = ["-tstamp"]


def execute(new_table, old_table, tz=None):  # pylint: disable=invalid-name
    for old_data in old_table.objects.all():
        new_data = new_table()
        if tz:
            new_data.__setattr__("timezone", tz)
        for field in old_data._meta.fields:
            field_name = field.name
            # 判断是否为外键
            if field_name in ("crontab", "interval", "worker"):
                try:
                    # 写入外键id
                    new_data.__setattr__(
                        field_name + "_id", old_data.__getattribute__(field_name).id
                    )
                except AttributeError:
                    new_data.__setattr__(field_name + "_id", None)
            else:
                new_data.__setattr__(field_name, old_data.__getattribute__(field_name))
        new_data.save()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from django.conf import settings
from django.contrib.auth.backends import ModelBackend

from blueapps import metrics
from blueapps.account import get_user_model
from blueapps.account.conf import ConfFixture
from blueapps.account.utils.http import send
from blueapps.utils import client
from blueapps.utils.tools import resolve_login_url

logger = logging.getLogger("component")

TOKEN_TYPE = "uin_skey"


class PtloginBackend(ModelBackend):
    def authenticate(self, request=None, uin=None, skey=None):

        if uin and skey:
            result = self.verify_uin_skey(uin, skey, request)
            if not result:
                return None
        else:
            return None

        user_model = get_user_model()
        user, _ = user_model.objects.get_or_create(username=uin)
        user_info = self.get_user_info(request, uin, skey)
        if user_info is not None:
            user.nickname = user_info["nick_name"]
            user.avatar_url = user_info["avatar_url"]
            user.save()
        return user

    @staticmethod
    def get_user_info(request, uin, skey):
        api_params = {"uin": uin, "skey": skey}
        # 根据ConfFixture.GET_USER_INFO_COLLECTION配置项获取esb调用对象
        get_user_info_collection = getattr(client, ConfFixture.GET_USER_INFO_COLLECTION)
        # 根据ConfFixture.GET_USER_INFO_COLLECTION_API配置项获取esb调用对象的api函数
        get_user_info_api = getattr(get_user_info_collection, ConfFixture.GET_USER_INFO_COLLECTION_API)
        # 根据GET_USER_INFO_COLLECTION_KWARGS配置项构造esb调用对象的api函数需要的参数
        get_user_info_params = {_key: api_params[_key] for _key in ConfFixture.GET_USER_INFO_COLLECTION_PARAMS}
        # 调用esb
        response = get_user_info_api(get_user_info_params)
        if response["result"]:
            return response["data"]
        else:
            return None

    @staticmethod
    def verify_uin_skey(uin, skey, request=None):
        """
        验证登录票据的有效性
        """
        api_params = {"app_code": settings.APP_CODE, "skey": skey, "uin": uin}
        # noinspection PyBroadException
        try:
            with metrics.observe(
                metrics.BLUEAPPS_USER_TOKEN_VERIFY_DURATION, hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE
            ):
                response = send(
                    resolve_login_url(ConfFixture.VERIFY_URL, request, "http"),
                    "GET",
                    api_params,
                )
            if not response["result"]:
                metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                    hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="verify_fail"
                ).inc()
                logger.error("uin，skey 验证失败，error=%s" % response["message"])
                return False
            else:
                return True
        except Exception as err:  # pylint: disable=broad-except
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="unknow_execption_raise"
            ).inc()
            logger.exception("uin，skey 验证异常，error=%s" % err)
            return False



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from __future__ import unicode_literals

import logging

from django.conf import settings
from django.utils.encoding import force_text
from django.utils.translation import gettext_lazy as _


class BlueException(Exception):
    """
    异常统一封装类
    """

    PLATFORM_CODE = getattr(settings, "PLATFORM_CODE", "00")
    MODULE_CODE = "00"
    ERROR_CODE = "500"
    MESSAGE = _("APP异常")
    STATUS_CODE = 500
    LOG_LEVEL = logging.ERROR

    def __init__(self, *args, message=None, data=None, **kwargs):
        """
        :param message: 错误消息
        :param data: 其他数据
        :param context: 错误消息 format dict
        :param args: 其他参数
        """
        super(BlueException, self).__init__(*args)

        if len(self.ERROR_CODE) == 3:
            code = f"{self.PLATFORM_CODE}{self.MODULE_CODE}{self.ERROR_CODE}"
        else:
            # 兼容旧版ERROR_CODE异常码长度不为3位 -> 只组合平台码和异常码
            code = f"{self.PLATFORM_CODE}{self.ERROR_CODE}"

        self.code = code
        self.errors = kwargs.get("errors")
        # 优先使用第三方系统的错误编码
        if kwargs.get("code"):
            self.code = kwargs["code"]

        self.message = force_text(message) if message else force_text(self.MESSAGE)

        self.data = data

    def __str__(self):
        return "[{}] {}".format(self.code, self.message)

    def render_data(self):
        """
        返回异常信息
        """
        return self.data

    def response_data(self):
        """
        返回response数据
        """
        return {
            "result": False,
            "code": self.code,
            "message": self.message,
            "data": self.render_data(),
        }


class ClientBlueException(BlueException):
    """
    客户端请求异常
    """

    MESSAGE = _("客户端请求异常")
    ERROR_CODE = "400"
    STATUS_CODE = 400


class ServerBlueException(BlueException):
    """
    服务器异常
    """

    MESSAGE = _("服务端服务异常")
    ERROR_CODE = "500"
    STATUS_CODE = 500


class ResourceNotFound(ClientBlueException):
    """
    客户端异常：找不到请求的资源
    """

    MESSAGE = _("找不到请求的资源")
    ERROR_CODE = "404"
    STATUS_CODE = 404


class ParamValidationError(ClientBlueException):
    """
    客户端异常：参数验证失败
    """

    MESSAGE = _("参数验证失败")
    ERROR_CODE = "400"
    STATUS_CODE = 400


class ParamRequired(ClientBlueException):
    """
    客户端异常：关键参数缺失
    """

    MESSAGE = _("关键参数缺失")
    ERROR_CODE = "401"
    STATUS_CODE = 400


class AccessForbidden(ClientBlueException):
    """
    客户端异常：登陆失败
    """

    MESSAGE = _("登陆失败")
    ERROR_CODE = "413"
    STATUS_CODE = 403


class RequestForbidden(ClientBlueException):
    """
    客户端异常：请求拒绝
    """

    MESSAGE = _("请求拒绝")
    ERROR_CODE = "423"
    STATUS_CODE = 403


class ResourceLock(ClientBlueException):
    """
    客户端异常：请求资源被锁定
    """

    MESSAGE = _("请求资源被锁定")
    ERROR_CODE = "433"
    STATUS_CODE = 403


class MethodError(ClientBlueException):
    """
    客户端异常：请求方法不支持
    """

    MESSAGE = _("请求方法不支持")
    ERROR_CODE = "405"
    STATUS_CODE = 405


class RioVerifyError(ClientBlueException):
    """
    客户端异常：智能网关检测失败
    """

    MESSAGE = _("登陆请求经智能网关检测失败")
    ERROR_CODE = "415"
    STATUS_CODE = 405


class BkJwtVerifyError(ClientBlueException):
    """
    客户端异常：JWT检测失败
    """

    MESSAGE = _("登陆请求经JWT检测失败")
    ERROR_CODE = "425"
    STATUS_CODE = 401


class DatabaseError(ServerBlueException):
    """
    服务端异常：数据库异常
    """

    MESSAGE = _("数据库异常")
    ERROR_CODE = "501"


class ApiNetworkError(ServerBlueException):
    """
    服务端异常：网络异常
    """

    MESSAGE = _("网络异常导致远程服务失效")
    ERROR_CODE = "503"
    STATUS_CODE = 503


class ApiResultError(ServerBlueException):
    """
    服务端异常：远程服务请求结果异常
    """

    MESSAGE = _("远程服务请求结果异常")
    ERROR_CODE = "513"
    STATUS_CODE = 503


class ApiNotAcceptable(ServerBlueException):
    """
    服务端异常：远程服务返回结果格式异常
    """

    MESSAGE = _("远程服务返回结果格式异常")
    ERROR_CODE = "523"
    STATUS_CODE = 503


class LocalError(ServerBlueException):
    """
    服务端异常：组件调用异常
    """

    MESSAGE = _("组件调用异常")
    ERROR_CODE = "533"


class LockError(ServerBlueException):
    MESSAGE = _("获取锁失败")
    ERROR_CODE = "543"



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from blueapps.account.conf import ConfFixture
from blueapps.utils import client
from blueapps.utils.esbclient import CustomComponentAPI


"""
发送短信工具文件，开发者可以直接调用此处的send_sms函数，屏蔽环境之间的差异
"""


def send_sms(user_list, content):
    """
    发送短信给指定的用户，
    :param user_list: 用户列表，list
    :param content: 消息内容
    :return: True | raise Exception
    """

    # 1. 获取发送短信的函数实际句柄
    sms_module = client.__getattr__(ConfFixture.SMS_CLIENT_MODULE)
    sms_func = sms_module.__getattr__(ConfFixture.SMS_CLIENT_FUNC)

    # 2. 拼接发送函数的内容
    request_args = {
        ConfFixture.SMS_CLIENT_USER_ARGS_NAME: ",".join(user_list),
        ConfFixture.SMS_CLIENT_CONTENT_ARGS_NAME: content,
    }

    # 3. 发送短信
    if type(sms_func) == CustomComponentAPI:
        result = sms_func.post(request_args)

    else:
        result = sms_func(request_args)

    return result



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


class ConfFixture(object):
    """
    登录模块项目变量汇总
    """

    #################
    # 浏览器参数说明 #
    #################

    # 登录模块,可选项为 components 目录下的模块
    BACKEND_TYPE = None

    # 用户验证 Backend
    USER_BACKEND = None

    # 用户登录验证中间件
    LOGIN_REQUIRED_MIDDLEWARE = None

    # 用户模型
    USER_MODEL = None

    # 登录平台链接
    LOGIN_URL = None

    # 内嵌式的登录平台链接（可嵌入弹框、IFrame）http://xxx.com/plain/
    LOGIN_PLAIN_URL = None

    # 是否提供内嵌式的统一登录页面
    HAS_PLAIN = True

    # 跳转至登录平台是否加跨域前缀标识
    # http://xxx.com/login/?c_url={CROSS_PREFIX}http%3A//xxx.com%3A8000/
    ADD_CROSS_PREFIX = True
    CROSS_PREFIX = ""

    # 跳转至登录平台是否加上APP_CODE
    # http://xxx.com/login/?c_url=http%3A//xxx.com%3A8000/&app_code=xxx
    ADD_APP_CODE = True
    # http://xxx.com/login/?c_url=http%3A//xxx.com%3A8000/&{APP_KEY}=xxx
    APP_KEY = "app_code"
    SETTINGS_APP_KEY = "APP_CODE"

    # 跳转至登录平台，回调参数名称
    # http://xxx.com/login/?{C_URL}=http%3A//xxx.com%3A8000/
    C_URL = "c_url"

    # 内嵌式的登录平台的尺寸大小，决定前端适配的弹框大小
    IFRAME_HEIGHT = 490
    IFRAME_WIDTH = 460

    ###############
    # 微信参数说明 #
    ###############

    # 登录模块 weixin
    WEIXIN_BACKEND_TYPE = None

    # 用户认证中间件 bk_ticket.middlewares.LoginRequiredMiddleware
    WEIXIN_MIDDLEWARE = None

    # 用户认证 Backend bk_ticket.backends.TicketBackend
    WEIXIN_BACKEND = None

    # 用户信息链接 http://xxx.com/user/weixin/get_user_info/
    WEIXIN_INFO_URL = None

    # 用户 OAUTH 认证链接 https://xxx.com/connect/oauth2/authorize
    WEIXIN_OAUTH_URL = None

    # 在微信端的应用ID 'xxxx'
    WEIXIN_APP_ID = None



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from django.contrib.auth.backends import ModelBackend

from blueapps import metrics
from blueapps.account import get_user_model
from blueapps.account.conf import ConfFixture
from blueapps.account.utils.http import send
from blueapps.utils.tools import resolve_login_url

logger = logging.getLogger("component")

TOKEN_TYPE = "weixin"


class WeixinBackend(ModelBackend):
    def authenticate(self, request=None, code=None, is_wechat=True):
        """
        is_wechat 参数是为了使得 WeixinBackend 与其他 Backend 参数个数不同，在框架选择
        认证 backend 时，快速定位
        """
        logger.debug("进入 WEIXIN 认证 Backend")
        if not code:
            return None

        result, user_info = self.verify_weixin_code(code, request)
        logger.debug("微信 CODE 验证结果，result：{}，user_info：{}".format(result, user_info))

        if not result:
            return None

        user_model = get_user_model()
        try:
            user, _ = user_model.objects.get_or_create(username=user_info["username"])
            user.nickname = user_info["username"]
            user.avatar_url = user_info["avatar"]
            user.save()
        except Exception:  # pylint: disable=broad-except
            logger.exception("自动创建 & 更新 User Model 失败")
        else:
            return user

    def get_user(self, user_id):
        user_model = get_user_model()
        try:
            return user_model.objects.get(pk=user_id)
        except user_model.DoesNotExist:
            return None

    @staticmethod
    def verify_weixin_code(code, request=None):
        """
        验证 WEIXIN 认证返回的授权码
        @param {string} code WEIXIN 认证返回的授权码
        @return {tuple} ret
        @return {boolean} ret[0] 是否认证通过
        @return {dict} ret[1] 当 result=True，该字段为用户信息，举例
            {
                u'username': u'',
                u'avatar': u''
            }
        """
        api_params = {
            "code": code,
        }
        try:
            with metrics.observe(
                metrics.BLUEAPPS_USER_TOKEN_VERIFY_DURATION, hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE
            ):
                response = send(
                    resolve_login_url(ConfFixture.WEIXIN_INFO_URL, request, "http"),
                    "GET",
                    api_params,
                )
            ret = response.get("ret")
            if ret == 0:
                return True, response["data"]
            else:
                metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                    hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="verify_fail"
                ).inc()
                logger.error("通过微信授权码，获取用户信息失败，error={}，ret={}".format(response["msg"], ret))
                return False, None
        except Exception:  # pylint: disable=broad-except
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="unknow_execption_raise"
            ).inc()
            logger.exception("通过微信授权码，获取用户信息异常")
            return False, None



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

try:
    from functools import wraps
except ImportError:
    from django.utils.functional import wraps  # Python 2.4 fallback.


# ===============================================================================
# 转义装饰器
# ===============================================================================
def escape_exempt(view_func):
    """
    转义豁免，被此装饰器修饰的action可以不进行中间件escape
    """

    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)

    wrapped_view.escape_exempt = True
    return wraps(view_func)(wrapped_view)


def escape_script(view_func):
    """
    被此装饰器修饰的action会对GET与POST参数进行javascript escape
    """

    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)

    wrapped_view.escape_script = True
    return wraps(view_func)(wrapped_view)


def escape_url(view_func):
    """
    被此装饰器修饰的action会对GET与POST参数进行url escape
    """

    def wrapped_view(*args, **kwargs):
        return view_func(*args, **kwargs)

    wrapped_view.escape_url = True
    return wraps(view_func)(wrapped_view)


def escape_exempt_param(*param_list, **param_list_dict):
    """
    此装饰器用来豁免某个view函数的某个参数
    @param param_list: 参数列表
    @return:
    """

    def _escape_exempt_param(view_func):
        def wrapped_view(*args, **kwargs):
            return view_func(*args, **kwargs)

        if param_list_dict.get("param_list"):
            wrapped_view.escape_exempt_param = param_list_dict["param_list"]
        else:
            wrapped_view.escape_exempt_param = list(param_list)
        return wraps(view_func)(wrapped_view)

    return _escape_exempt_param



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import abc
import sys
from enum import Enum
from typing import Tuple


def inject_logging_trace_info(
    logging: dict, inject_formatters: Tuple[str], trace_format: str, format_keywords: Tuple[str] = ("format", "fmt")
):
    """往logging配置中动态注入trace信息，直接修改logging数据"""
    formatters = {name: formatter for name, formatter in logging["formatters"].items() if name in inject_formatters}
    for name, formatter in formatters.items():
        matched_keywords = set(format_keywords).intersection(set(formatter.keys()))
        for keyword in matched_keywords:
            formatter.update({keyword: formatter[keyword].strip() + f" {trace_format}\n"})


class BaseServiceNameHandler(abc.ABC):
    def __init__(self, service_name: str):
        self._service_name = service_name

    @abc.abstractmethod
    def get_service_name(self) -> str:
        ...


class BlueappsServiceNameHandler(BaseServiceNameHandler):
    class SuffixEnum(Enum):
        CELERY_BEAT = "celery_beat"
        CELERY_WORKER = "celery_worker"
        API = "api"
        NONE = ""

    @property
    def is_celery(self):
        return "celery" in sys.argv

    @property
    def is_celery_beat(self):
        return self.is_celery and "beat" in sys.argv

    @property
    def suffix(self):
        if self.is_celery_beat:
            return self.SuffixEnum.CELERY_BEAT.value
        if self.is_celery:
            return self.SuffixEnum.CELERY_WORKER.value
        return self.SuffixEnum.NONE.value

    def get_service_name(self) -> str:
        if self.suffix:
            return "{}_{}".format(self._service_name, self.suffix)
        return self._service_name



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.conf import settings
from django.utils.module_loading import import_string

# 变量默认配置
DEFAULT_SETTINGS = {
    # ACCOUNT 相关配置会根据有无配置进行值覆盖，默认不提供，默认值可参考account/sites/default.py
    "ENABLE": {
        "OTEL": {
            "TRACE": False,
            "METRICS": False,
        }
    },
    "BKAPP": {
        "OTEL": {
            "SERVICE_NAME": getattr(settings, "APP_CODE", "Blueapps"),
            "BK_DATA_ID": -1,
            "SAMPLER": "parentbased_always_off",
            "GRPC_HOST": None,
            "BK_DATA_TOKEN": "",
            "LOGGING_TRACE_FORMAT": None,
            "INSTRUMENT_DB_API": False,
            "ADDTIONAL_INSTRUMENTORS": [],
            "SPAN_PROCESSOR": "blueapps.opentelemetry.export.LazyBatchSpanProcessor",
            "SERVICE_NAME_HANDLER": "blueapps.opentelemetry.utils.BlueappsServiceNameHandler",
        },
    },
    "IS_AJAX_PLAIN_MODE": False,
    "SPECIFIC_REDIRECT_KEY": None,
    "AJAX_401_RESPONSE_FUNC": None,
    "PAGE_401_RESPONSE_FUNC": None,
    "PAGE_401_RESPONSE_PLATFORM_FUNC": None,
    "NON_REQUEST_USERNAME_PROVIDER": None,
}

# 提供路径需要import_string的变量
IMPORT_SETTINGS = [
    "AJAX_401_RESPONSE_FUNC",
    "PAGE_401_RESPONSE_FUNC",
    "PAGE_401_RESPONSE_PLATFORM_FUNC",
    "BKAPP_OTEL_SPAN_PROCESSOR",
    "BKAPP_OTEL_SERVICE_NAME_HANDLER",
    "NON_REQUEST_USERNAME_PROVIDER",
]

# 在django_settings和blueapps_settings中都支持的配置项
BLUEAPPS_SUPPORT_DJANGO_SETTINGS = [
    "IS_AJAX_PLAIN_MODE",
    "ENABLE_OTEL_TRACE",
    "ENABLE_OTEL_METRICS",
    "BKAPP_OTEL_SERVICE_NAME",
    "BKAPP_OTEL_BK_DATA_ID",
    "BKAPP_OTEL_SAMPLER",
    "BKAPP_OTEL_GRPC_HOST",
    "BKAPP_OTEL_BK_DATA_TOKEN",
    "BK_APP_OTEL_ADDTIONAL_INSTRUMENTORS",
    "BK_APP_OTEL_INSTRUMENT_DB_API",
    "OTEL_LOGGING_TRACE_FORMAT",
    "BKAPP_OTEL_SPAN_PROCESSOR",
    "BKAPP_OTEL_SERVICE_NAME_HANDLER",
]


class BlueappsSettings:
    SETTING_PREFIX = "BLUEAPPS"
    NESTING_SEPARATOR = "_"

    def __init__(self, default_settings=None, import_strings=None):
        self.project_settings = self.get_flatten_settings(getattr(settings, self.SETTING_PREFIX, {}))
        self.project_settings.update(self.get_django_blueapps_settings())
        self.default_settings = self.get_flatten_settings(default_settings or DEFAULT_SETTINGS)
        self.import_strings = import_strings or IMPORT_SETTINGS

    def __getattr__(self, key):
        if key not in self.project_settings and key not in self.default_settings:
            raise AttributeError

        try:
            value = self.project_settings[key]
        except KeyError:
            value = self.default_settings[key]

        if key in self.import_strings and isinstance(value, str):
            try:
                value = import_string(value)
            except ImportError as e:
                message = f"Can not import {value} for Blueapps settings: {e}"
                raise ImportError(message)

        if value is not None:
            setattr(self, key, value)
        return value

    def get_flatten_settings(self, inputted_settings: dict, cur_prefix: str = ""):
        """获取BLUEAPPS配置字典打平之后的配置字典"""

        def get_cur_key(cur_key):
            return f"{cur_prefix}{self.NESTING_SEPARATOR}{cur_key}" if cur_prefix else cur_key

        flatten_settings = {}
        for key, value in inputted_settings.items():
            if isinstance(value, dict):
                flatten_sub_settings = self.get_flatten_settings(value, key)
                flatten_settings.update(
                    {
                        get_cur_key(flatten_key): flatten_value
                        for flatten_key, flatten_value in flatten_sub_settings.items()
                    }
                )
            else:
                flatten_settings[get_cur_key(key)] = value
        return flatten_settings

    def get_django_blueapps_settings(self):
        """获取分散配置在settings中配置项"""
        django_setting_keys = [
            key
            for key in dir(settings)
            if (
                key.startswith(f"{self.SETTING_PREFIX}{self.NESTING_SEPARATOR}")
                or key in BLUEAPPS_SUPPORT_DJANGO_SETTINGS
            )
        ]
        prefix_len = len(f"{self.SETTING_PREFIX}{self.NESTING_SEPARATOR}")
        return {
            key if key in BLUEAPPS_SUPPORT_DJANGO_SETTINGS else key[prefix_len:]: getattr(settings, key)
            for key in django_setting_keys
        }


blueapps_settings = BlueappsSettings(DEFAULT_SETTINGS, IMPORT_SETTINGS)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import hashlib
import logging
import time

from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.utils.translation import gettext_lazy as _

from blueapps import metrics
from blueapps.account import get_user_model

logger = logging.getLogger("component")

TOKEN_TYPE = "rio"


class RioBackend(ModelBackend):
    def authenticate(
        self,
        request=None,
        timestamp=None,
        signature=None,
        staff_id=None,
        staff_name=None,
        x_ext_data=None,
        x_rio_seq=None,
    ):
        logger.debug("进入 RIO 认证 Backend")

        try:
            verify_data = self.verify_rio_request(
                request,
                timestamp=timestamp,
                signature=signature,
                staff_id=staff_id,
                staff_name=staff_name,
                x_ext_data=x_ext_data,
                x_rio_seq=x_rio_seq,
            )
        except UnicodeEncodeError as error:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="rio_encode_error"
            ).inc()
            logger.exception("[RIO]编码异常: %s" % error)
            return None
        except Exception as err:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="unknow_execption_raise"
            ).inc()
            logger.exception("[RIO]校验异常: %s" % err)
            return None

        logger.debug("RIO 验证结果，verify_data：%s" % verify_data)

        if not verify_data["result"] or not verify_data["data"]:
            return None

        user_info = verify_data["data"]
        user_model = get_user_model()
        try:
            user, _ = user_model.objects.get_or_create(username=user_info["username"])
            user.nickname = user_info["username"]
            user.save()
        except Exception as err:  # pylint: disable=broad-except
            logger.exception("自动创建 & 更新 User Model 失败: %s" % err)
            return None

        return user

    def get_user(self, user_id):
        user_model = get_user_model()
        try:
            return user_model.objects.get(pk=user_id)
        except user_model.DoesNotExist:
            return None

    @staticmethod
    def verify_rio_request(request, timestamp, signature, staff_id, staff_name, x_ext_data, x_rio_seq):
        """
        验证 RIO请求
        @param {string} timestamp 当前Unix时间戳，精确到秒，用于判断当前签名的时效性
        @param {string} signature 当前请求的签名(大写英文)，用于判断当前请求的合法
        @param {string} staff_id 员工的id
        @param {string} staff_name 员工的英文名
        @param {string} x_ext_data 微信的请求该字段为用户的 OpenID
        @param {string} x_rio_seq 当前请求的唯一的标识
        @return {dict} ret[1] 当 result=True，该字段为用户信息，举例
            {
                u'username': u'',
                u'avatar': u''
            }
        """
        ret = {"result": False, "message": "", "data": {}}

        if not timestamp or not signature:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="param_loss"
            ).inc()
            logger.error("[RIO] login exception, param loss")
            ret["message"] = _("[RIO]login exception")
            return ret

        now = int(time.time())
        if abs(now - int(timestamp)) > 180:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="rio_expired"
            ).inc()
            logger.error("[RIO] login time expire")
            ret["message"] = _("[RIO]login time expire")
            return ret

        token = getattr(settings, "RIO_TOKEN")

        # 校验
        with metrics.observe(
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_DURATION, hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE
        ):
            plaintext = timestamp + token + x_rio_seq + "," + staff_id + "," + staff_name + "," + x_ext_data + timestamp

            sign = hashlib.sha256(plaintext.encode()).hexdigest()

            if signature.lower() != sign.lower():
                logger.error("[RIO] login signature error {} vs {}".format(sign, signature))
                ret["message"] = _("[RIO]login signature error")
                return ret

        # 通过TOF获取用户信息
        # staff_info = get_staff_info_by_login_name(staffname)
        # if not staff_info:
        #    logger.error(u"[RIO]通过TOF获取用户信息异常")
        #    return result, data
        ret["result"] = True
        ret["data"] = {
            "username": staff_name,
        }
        return ret



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import json
from typing import Collection

from opentelemetry.instrumentation import dbapi
from opentelemetry.instrumentation.celery import CeleryInstrumentor
from opentelemetry.instrumentation.django import DjangoInstrumentor
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.logging import LoggingInstrumentor
from opentelemetry.instrumentation.redis import RedisInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.trace import Span, Status, StatusCode, format_trace_id
from requests import Response

from blueapps.settings import blueapps_settings


def requests_callback(span: Span, response: Response):
    """
    处理蓝鲸标准协议响应
    """
    try:
        json_result = response.json()
    except Exception:  # pylint: disable=broad-except
        return

    if not isinstance(json_result, dict):
        return

    # NOTE: esb has a result=bool, but apigateway or other backend maybe has no result
    code = json_result.get("code", 0)
    try:
        code = int(code)
    except Exception:  # pylint: disable=broad-except
        pass

    span.set_attribute("result_code", code)
    span.set_attribute("result_message", json_result.get("message", ""))

    errors = str(json_result.get("errors", ""))
    if errors:
        span.set_attribute("result_errors", errors)

    request_id = (
        # new esb and apigateway
        response.headers.get("x-bkapi-request-id")
        # iam backend
        or response.headers.get("x-request-id")
        # old esb
        or json_result.get("request_id", "")
    )
    if request_id:
        span.set_attribute("request_id", request_id)

    span.set_status(Status(StatusCode.OK if response.ok else StatusCode.ERROR))


def django_request_hook(span, request):
    """
    在request注入trace_id，方便获取
    """
    trace_id = span.get_span_context().trace_id
    request.otel_trace_id = format_trace_id(trace_id)


def django_response_hook(span, request, response):
    """
    处理蓝鲸标准协议 Django 响应
    """
    if hasattr(response, "data"):
        result = response.data
    else:
        try:
            result = json.loads(response.content)
        except Exception:  # pylint: disable=broad-except
            return
    if not isinstance(result, dict):
        return
    span.set_attribute("result_code", result.get("code", 0))
    span.set_attribute("result_message", result.get("message", ""))
    span.set_attribute("result_errors", result.get("errors", ""))
    result = result.get("result", True)
    if result:
        span.set_status(Status(StatusCode.OK))
        return
    span.set_status(Status(StatusCode.ERROR))


class BKAppInstrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self) -> Collection[str]:
        return []

    def _instrument(self, **kwargs):
        LoggingInstrumentor().instrument()
        RequestsInstrumentor().instrument(span_callback=requests_callback)
        DjangoInstrumentor().instrument(request_hook=django_request_hook, response_hook=django_response_hook)
        CeleryInstrumentor().instrument()
        RedisInstrumentor().instrument()

        for instrumentor in blueapps_settings.BK_APP_OTEL_ADDTIONAL_INSTRUMENTORS or getattr(
            blueapps_settings, "BK_APP_OTEL_ADDTIONAL_INSTRUMENTORS", []
        ):
            instrumentor.instrument()

        if blueapps_settings.BKAPP_OTEL_INSTRUMENT_DB_API or getattr(
            blueapps_settings, "BK_APP_OTEL_INSTRUMENT_DB_API", False
        ):
            import MySQLdb  # noqa

            dbapi.wrap_connect(
                __name__,
                MySQLdb,
                "connect",
                "mysql",
                {"database": "db", "port": "port", "host": "host", "user": "user"},
            )

    def _uninstrument(self, **kwargs):
        for instrumentor in self.instrumentors:
            instrumentor.uninstrument()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

# Generated by Django 1.11.2 on 2017-08-09 14:51
from __future__ import unicode_literals

import django.core.validators
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0008_alter_user_username_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="last login"
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={
                            "unique": "A user with that openid already exists."
                        },
                        help_text="Required. 64 characters or fewer. Letters, digits and underlined only.",
                        max_length=64,
                        unique=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^[a-zA-Z0-9_]+$",
                                "Enter a valid openid. This value may contain only letters, "
                                "numbers and underlined characters.",
                                "invalid",
                            )
                        ],
                        verbose_name="username",
                    ),
                ),
                (
                    "nickname",
                    models.CharField(
                        blank=True,
                        help_text="Required. 64 characters or fewer.",
                        max_length=64,
                        verbose_name="nick name",
                    ),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. "
                                  "Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. "
                                  "A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={"verbose_name": "user", "verbose_name_plural": "users"},
        ),
        migrations.CreateModel(
            name="UserProperty",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "key",
                    models.CharField(
                        help_text="Required. 64 characters or fewer. Letters, digits and underlined only.",
                        max_length=64,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^[a-zA-Z0-9_]+$",
                                "Enter a valid key. This value may contain only letters, "
                                "numbers and underlined characters.",
                                "invalid",
                            )
                        ],
                    ),
                ),
                ("value", models.TextField()),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="properties",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "db_table": "account_user_property",
                "verbose_name": "user property",
                "verbose_name_plural": "user properties",
            },
        ),
        migrations.CreateModel(
            name="UserProxy",
            fields=[],
            options={"proxy": True, "indexes": [], },
            bases=("account.user",),
        ),
        migrations.AlterUniqueTogether(name="userproperty", unique_together={("user", "key")}),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.


BK JWT - settings option
1. settings.APIGW_ENABLED: if enabled, it will automatically get the apigw public key
2. settings.APIGW_PUBLIC_KEY: If you get the apigw public key manually
3. settings.APIGW_API_ACCOUNT: The account used to call the apigw API
4. settings.APIGW_JWT_KEY: Apigw uses this key to pass jwt encrypted content
"""

import logging

from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.utils.translation import ugettext_lazy as _

from blueapps import metrics
from blueapps.account import get_user_model
from blueapps.utils.jwt_client import JWTClient

logger = logging.getLogger("component")  # pylint: disable=invalid-name

TOKEN_TYPE = "bk_jwt"


class BkJwtBackend(ModelBackend):
    def authenticate(self, request=None):
        logger.debug("进入 BK_JWT 认证 Backend")

        try:
            verify_data = self.verify_bk_jwt_request(request)
        except Exception as err:  # pylint: disable=broad-except
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="unknow_execption_raise"
            ).inc()
            logger.exception("[BK_JWT]校验异常: %s" % err)
            return None

        if not verify_data["result"] or not verify_data["data"]:
            logger.error("BK_JWT 验证失败： %s" % verify_data)
            return None

        user_info = verify_data["data"]["user"]
        user_model = get_user_model()
        try:
            user, _ = user_model.objects.get_or_create(
                defaults={"nickname": user_info["bk_username"]},
                username=user_info["bk_username"],
            )
        except Exception as err:  # pylint: disable=broad-except
            logger.exception("自动创建 & 更新 User Model 失败: %s" % err)
            return None

        return user

    def get_user(self, user_id):
        user_model = get_user_model()
        try:
            return user_model.objects.get(pk=user_id)
        except user_model.DoesNotExist:
            return None

    @staticmethod
    def verify_bk_jwt_request(request):
        """
        验证 BK_JWT 请求
        @param {string} x_bkapi_jwt JWT请求头
        @return {dict}
            {
                'result': True,
                'message': '',
                'data': {
                    'user': {
                        'bk_username': '调用方用户'
                    },
                    'app': {
                        'bk_app_code': '调用方app'
                    }
                }
            }
        """
        ret = {"result": False, "message": "", "data": {}}

        with metrics.observe(
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_DURATION, hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE
        ):
            jwt = JWTClient(request)

        if not jwt.is_valid:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="invalid_jwt"
            ).inc()
            ret["message"] = _("jwt_invalid: %s") % jwt.error_message
            return ret

        # verify: user && app
        app = jwt.get_app_model()
        if not app["verified"]:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="app_verify_err"
            ).inc()
            ret["message"] = app.get("valid_error_message", _("APP鉴权失败"))
            ret["data"]["app"] = app
            return ret

        if not app.get("bk_app_code"):
            app["bk_app_code"] = app["app_code"]

        user = jwt.get_user_model()
        # ESB默认需要校验用户信息
        use_esb_white_list = getattr(settings, "USE_ESB_WHITE_LIST", True)

        if not use_esb_white_list and not user["verified"]:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="user_verify_err"
            ).inc()
            ret["message"] = user.get("valid_error_message", _("用户鉴权失败且不支持ESB白名单"))
            ret["data"]["user"] = user
            return ret
        if not user.get("bk_username"):
            user["bk_username"] = user["username"]

        if not app["bk_app_code"]:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="empty_app_code"
            ).inc()
            ret["message"] = _("无法获取bk_app_code")
            return ret

        if not user["bk_username"]:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="empty_user"
            ).inc()
            ret["message"] = _("无法获取用户信息")
            return ret

        ret["result"] = True
        ret["data"] = {"user": user, "app": app}
        return ret



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from __future__ import unicode_literals

import datetime
import logging
import random
import traceback

from django.conf import settings
from django.contrib.auth.models import (
    AbstractBaseUser,
    BaseUserManager,
    PermissionsMixin,
)
from django.core import validators
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from blueapps.account import conf
from blueapps.account.utils import sms
from blueapps.utils.crypto import SymmetricTextField

ConfFixture = conf.ConfFixture

# 结合获取用户配置的二步验证配置结果
SV_CONF = conf.SECOND_VERIFY_CONF
user_sv_conf = getattr(settings, "SECOND_VERIFY_CONF", {})
SV_CONF.update(user_sv_conf)

logger = logging.getLogger("app")


class UserManager(BaseUserManager):
    def _create_user(self, username, is_staff=False, is_superuser=False, password=None, **extra_fields):
        now = timezone.now()
        if not username:
            raise ValueError(_("The given username must be set"))
        user = self.model(
            username=username,
            is_active=True,
            is_staff=is_staff,
            is_superuser=is_superuser,
            date_joined=now,
            **extra_fields
        )
        if password:
            user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, username, password=None, **extra_fields):
        return self._create_user(username, False, False, password, **extra_fields)

    def create_superuser(self, username, password=None, **extra_fields):
        return self._create_user(username, True, True, password, **extra_fields)


class User(AbstractBaseUser, PermissionsMixin):
    username = models.CharField(
        _("username"),
        max_length=64,
        unique=True,
        help_text=_("Required. 64 characters or fewer. Letters, " "digits and underlined only."),
        validators=[
            validators.RegexValidator(
                r"^[a-zA-Z0-9_]+$",
                _(
                    "Enter a valid openid. "
                    "This value may contain only letters, "
                    "numbers and underlined characters."
                ),
                "invalid",
            ),
        ],
        error_messages={"unique": _("A user with that openid already exists.")},
    )

    nickname = models.CharField(
        _("nick name"),
        max_length=64,
        blank=True,
        help_text=_("Required. 64 characters or fewer."),
    )
    is_staff = models.BooleanField(
        _("staff status"),
        default=False,
        help_text=_("Designates whether the user can log into this " "admin site."),
    )
    is_active = models.BooleanField(
        _("active"),
        default=True,
        help_text=_(
            "Designates whether this user should be treated as " "active. Unselect this instead of deleting accounts."
        ),
    )
    date_joined = models.DateTimeField(_("date joined"), default=timezone.now)

    objects = UserManager()

    USERNAME_FIELD = "username"
    REQUIRED_FIELDS = ["nickname"]

    class Meta:
        verbose_name = _("user")
        verbose_name_plural = _("users")

        # Pass platform default user table
        # db_table = 'auth_user'

    def get_full_name(self):
        full_name = "{}({})".format(self.username, self.nickname)
        return full_name.strip()

    def get_short_name(self):
        return self.nickname

    def get_property(self, key):
        try:
            return self.properties.get(key=key).value
        except UserProperty.DoesNotExist:
            return None

    def set_property(self, key, value):
        key_property, _ = self.properties.get_or_create(key=key)
        key_property.value = value
        key_property.save()

    @property
    def avatar_url(self):
        return self.get_property("avatar_url")

    @avatar_url.setter
    def avatar_url(self, a_url):
        self.set_property("avatar_url", a_url)

    def send_sms(self, code):

        try:
            result = sms.send_sms([self.username], SV_CONF["SMS_FORMAT"].format(code))
        except Exception:  # pylint: disable=broad-except
            logger.error(
                "cmsi.send_sms_for_external_user failed. "
                "username->[%s], code->[%s] for->[%s]" % (self.username, code, traceback.format_exc())
            )
            return {"result": False, "message": _("ESB发送短信接口错误，可能由权限问题导致")}
        return {"result": result["result"], "message": result["message"]}

    def send_code(self):
        now = timezone.now()
        v_info = VerifyInfo.objects.filter(user=self)
        v_info_cnt = v_info.count()
        if v_info_cnt == 0:
            # 从未发送过验证码 或 发送过验证码但已被使用（使用后会删除db中的记录）
            # 生成新的验证码并发送
            code = random.randint(111111, 999999)
            VerifyInfo.objects.create(user=self, code=code)
            ret = self.send_sms(code)
            if ret["result"]:
                ret["message"] = _("初始化验证码，发送成功")

        elif v_info_cnt == 1:
            cur = v_info[0]
            if cur.updated_at >= now - datetime.timedelta(minutes=SV_CONF["VALID_MINUTES"]):
                # 早前生成过验证码，且未过期
                if cur.updated_at < now - datetime.timedelta(minutes=SV_CONF["RETRY_MINUTES"]):
                    # 重发已生成的
                    ret = self.send_sms(cur.code)
                    if ret["result"]:
                        ret["message"] = _("已生成的验证码，重发成功")
                else:
                    # 等待时间不足，不重发
                    ret = {"result": False, "message": _("暂不能重发验证码，请稍等")}
            else:
                # 已过期，重新生成并重发
                new_code = random.randint(111111, 999999)
                cur.code = new_code
                cur.save()
                ret = self.send_sms(new_code)
                if ret["result"]:
                    ret["message"] = _("重新生成验证码，发送成功")
        else:
            logger.error("found more than one code of the user->[%s]" % self.id)
            ret = {"result": False, "message": _("数据库中的验证码异常")}
        return ret

    def verify_code(self, code):
        check = VerifyInfo.objects.filter(
            user=self,
            code=code,
            updated_at__gt=timezone.now() - datetime.timedelta(minutes=SV_CONF["VALID_MINUTES"]),
        ).count()
        if check == 1:
            # 一个验证码只能用一次 用完删除
            VerifyInfo.objects.filter(user=self, code=code).delete()
            return True
        return False


class UserProperty(models.Model):
    """
    Add user extra property
    """

    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="properties")
    key = models.CharField(
        max_length=64,
        help_text=_("Required. 64 characters or fewer. Letters, " "digits and underlined only."),
        validators=[
            validators.RegexValidator(
                r"^[a-zA-Z0-9_]+$",
                _("Enter a valid key. " "This value may contain only letters, " "numbers and underlined characters."),
                "invalid",
            ),
        ],
    )
    value = SymmetricTextField(using="blueapps")

    class Meta:
        verbose_name = _("user property")
        verbose_name_plural = _("user properties")
        db_table = "account_user_property"
        unique_together = (("user", "key"),)


class VerifyInfo(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    code = models.CharField(max_length=6)
    updated_at = models.DateTimeField(auto_now=True)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from rest_framework.authentication import BasicAuthentication, SessionAuthentication

from blueapps.contrib.drf.serializer import custom_params_valid
from blueapps.utils.logger import logger


class ApiMixin:
    """
    封装 APIViewSet 修改 ModelViewSet 默认返回内容，固定格式为
        {result: True, data: {}, code: 00, message: ''}
    """

    authentication_classes = (BasicAuthentication, SessionAuthentication)

    def initialize_request(self, request, *args, **kwargs):
        """
        Set the `.action` attribute on the view, depending on the request method.
        """
        logger.info(
            "[receive request], path: {}, header: {}, body: {}".format(
                request.path, request.headers.get("X-Bkapi-App"), request.body
            )
        )
        return super(ApiMixin, self).initialize_request(request, *args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        """
        `.dispatch()` is pretty much the same as Django's regular dispatch,
        but with extra hooks for startup, finalize, and exception handling.
        """
        return super().dispatch(request, *args, **kwargs)


class ValidationMixin:
    """
    DRF视图自定义校验器
    """

    def params_valid(self, serializer, params=None, instance=None, many=False, partial=False):
        """
        校验参数是否满足 serializer 规定的格式，支持传入serializer
        """
        # 校验request中的参数
        if not params:
            if self.request.method in ["GET"]:
                params = self.request.query_params
            else:
                params = self.request.data

        return custom_params_valid(serializer=serializer, params=params, instance=instance, many=many, partial=partial,)

    @property
    def validated_data(self):
        """
        校验的数据
        """
        # 优先使用缓存，避免多次取值时重复序列化和校验
        cache_validated_data = getattr(self, "cache_validated_data", None)
        if cache_validated_data:
            return cache_validated_data

        if self.request.method == "GET":
            data = self.request.query_params
        else:
            data = self.request.data
        serializer = self.serializer_class or self.get_serializer_class()
        cache_validated_data = self.params_valid(serializer, data)
        setattr(self, "cache_validated_data", cache_validated_data)
        return cache_validated_data



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import errno
import os
import shutil
from os import path

import django
from django.core.management import CommandError
from django.core.management.templates import TemplateCommand
from django.core.management.utils import handle_extensions
from django.template import Context, Engine
from django.utils.version import get_docs_version

import blueapps


class BlueTemplateCommand(TemplateCommand):
    def handle_template(self, template, subdir):
        if template is None:
            return path.join(blueapps.__path__[0], "conf", subdir)

        else:
            return super(BlueTemplateCommand, self).handle_template(template, subdir)

    def handle(self, app_or_project, name, target=None, **options):
        self.app_or_project = app_or_project
        self.paths_to_remove = []
        self.verbosity = options["verbosity"]

        self.validate_name(name, app_or_project)
        # if some directory is given, make sure it's nicely expanded
        top_dir = self.get_top_dir(target, name)

        extensions = tuple(handle_extensions(options["extensions"]))
        extra_files = ["csrftoken.js"]
        for file in options["files"]:
            extra_files.extend(map(lambda x: x.strip(), file.split(",")))
        if self.verbosity >= 2:
            self.stdout.write(
                "Rendering %s template files with "
                "extensions: %s\n" % (app_or_project, ", ".join(extensions))
            )
            self.stdout.write(
                "Rendering %s template files with "
                "filenames: %s\n" % (app_or_project, ", ".join(extra_files))
            )

        base_name = "%s_name" % app_or_project
        base_subdir = "%s_template" % app_or_project
        base_directory = "%s_directory" % app_or_project

        context = Context(
            dict(
                options,
                **{
                    base_name: name,
                    base_directory: top_dir,
                    "docs_version": get_docs_version(),
                    "django_version": django.__version__,
                }
            ),
            autoescape=False,
        )

        # Setup a stub settings environment for template rendering
        from django.conf import settings

        if not settings.configured:
            settings.configure()

        template_dir = self.handle_template(options["template"], base_subdir)
        prefix_length = len(template_dir) + 1

        for root, dirs, files in os.walk(template_dir):

            path_rest = root[prefix_length:]
            relative_dir = path_rest.replace(base_name, name)
            if relative_dir:
                target_dir = path.join(top_dir, relative_dir)
                if not path.exists(target_dir):
                    os.mkdir(target_dir)

            for dirname in dirs[:]:
                if dirname.startswith(".") or dirname == "__pycache__":
                    dirs.remove(dirname)
                # 处理多版本差异，将只对指定版本初始化
                if "run_ver" in options and os.path.basename(root) == "sites":
                    if dirname != options["run_ver"]:
                        dirs.remove(dirname)
            self.remove_or_write_file(files, root, top_dir, relative_dir,
                                      base_name, name, extensions, extra_files, context)

        if self.paths_to_remove:
            if self.verbosity >= 2:
                self.stdout.write("Cleaning up temporary files.\n")
            for path_to_remove in self.paths_to_remove:
                if path.isfile(path_to_remove):
                    os.remove(path_to_remove)
                else:
                    shutil.rmtree(path_to_remove)

    def remove_or_write_file(self, files, root, top_dir, relative_dir,
                             base_name, name, extensions, extra_files,
                             context):
        for filename in files:
            if filename.endswith((".pyo", ".pyc", ".py.class")):
                # Ignore some files as they cause various breakages.
                continue
            old_path = path.join(root, filename)
            new_path = path.join(
                top_dir, relative_dir, filename.replace(base_name, name)
            )
            if path.exists(new_path):
                raise CommandError(
                    "%s already exists, overlaying a "
                    "project or app into an existing "
                    "directory won't replace conflicting "
                    "files" % new_path
                )

            # Only render the Python files, as we don't want to
            # accidentally render Django templates files
            with open(old_path, "rb") as template_file:
                content = template_file.read()
            if filename.endswith(extensions) or filename in extra_files:
                content = content.decode("utf-8")
                template = Engine().from_string(content)
                content = template.render(context)
                content = content.encode("utf-8")
            with open(new_path, "wb") as new_file:
                new_file.write(content)

            if self.verbosity >= 2:
                self.stdout.write("Creating %s\n" % new_path)
            try:
                shutil.copymode(old_path, new_path)
                self.make_writeable(new_path)
            except OSError:
                self.stderr.write(
                    "Notice: Couldn't set permission bits on %s. You're "
                    "probably using an uncommon filesystem setup. No "
                    "problem." % new_path,
                    self.style.NOTICE,
                )

    @staticmethod
    def get_top_dir(target, name):
        if target is None:
            top_dir = path.join(os.getcwd(), name)
            try:
                os.makedirs(top_dir)
            except OSError as err:
                if err.errno == errno.EEXIST:
                    message = "'%s' already exists" % top_dir
                else:
                    message = err
                raise CommandError(message)
        else:
            top_dir = os.path.abspath(path.expanduser(target))
            if not os.path.exists(top_dir):
                raise CommandError(
                    "Destination directory '%s' does not "
                    "exist, please create it first." % top_dir
                )
        return top_dir



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _
from django.conf import settings
from django.core.cache import caches

from blueapps.account.conf import APIGW_CACHE_KEY, APIGW_CACHE_EXPIRES

cache = caches["login_db"]


class AccountConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "blueapps.account"
    verbose_name = _("account")

    def ready(self):
        from .signals import user_logged_out_handler  # noqa

        self.get_api_public_key()
        return True

    @staticmethod
    def get_api_public_key():
        # return if APIGW_ENABLED is not set
        if not hasattr(settings, "APIGW_ENABLED") or not bool(settings.APIGW_ENABLED):
            return

        # return if public key is configured
        if hasattr(settings, "APIGW_PUBLIC_KEY"):
            return

        # get api key by cache
        from blueapps.utils.logger import logger  # noqa

        api_public_key = cache.get(APIGW_CACHE_KEY)
        if api_public_key:
            message = "[ESB][JWT]get esb api public key success (from cache)"
            print(message, flush=True)
            logger.info(message)
            settings.APIGW_PUBLIC_KEY = api_public_key
            return api_public_key

        # get api key by api
        from blueapps.utils import get_client_by_user  # noqa

        client = get_client_by_user(getattr(settings, "APIGW_API_ACCOUNT", "admin"))
        esb_result = client.esb.get_api_public_key()
        if esb_result["result"]:
            api_public_key = esb_result["data"]["public_key"]
            settings.APIGW_PUBLIC_KEY = api_public_key
            cache.set(APIGW_CACHE_KEY, api_public_key, APIGW_CACHE_EXPIRES)
            message = "[ESB][JWT]get esb api public key success (from realtime api)"
            print(message, flush=True)
            logger.info(message)
        else:
            message = f'[ESB][JWT]get esb api public key error:{esb_result["message"]}'
            print(message, flush=True)
            logger.warning(message)
            raise Exception(message)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from __future__ import absolute_import

import datetime
import json
import logging

from django.conf import settings

logger = logging.getLogger("blueapps")


def blue_settings(request):
    try:
        if request.user.is_anonymous:
            username = ""
            nickname = ""
            avatar_url = ""
        else:
            username = request.user.username
            nickname = request.user.nickname
            avatar_url = request.user.avatar_url

        context = {
            # 本地静态文件访问
            "STATIC_URL": settings.STATIC_URL,
            # 当前页面，主要为了login_required做跳转用
            "APP_PATH": request.get_full_path(),
            # 运行模式
            "RUN_MODE": settings.RUN_MODE,
            # 运行版本（内部版、混合云版...）
            "RUN_VER": settings.RUN_VER,
            # 在蓝鲸系统中注册的  "应用编码"
            "APP_CODE": settings.APP_CODE,
            # URL前缀
            "SITE_URL": settings.SITE_URL,
            # 远程静态资源url
            "REMOTE_STATIC_URL": settings.REMOTE_STATIC_URL,
            # 静态资源版本号,用于指示浏览器更新缓存
            "STATIC_VERSION": settings.STATIC_VERSION,
            # 蓝鲸平台URL
            "BK_URL": settings.BK_URL,
            # 蓝鲸开发者页面
            "BK_DEV_URL": settings.BK_DEV_URL,
            # 用户名
            "USERNAME": username,
            # 用户昵称
            "NICKNAME": nickname,
            # 用户头像
            "AVATAR_URL": avatar_url,
            # WEIXIN ROOT URL
            "WEIXIN_SITE_URL": settings.WEIXIN_SITE_URL,
            # WEIXIN 本地静态资源链接
            "WEIXIN_STATIC_URL": settings.WEIXIN_STATIC_URL,
            # WEIXIN 远程静态资源链接
            "WEIXIN_REMOTE_STATIC_URL": settings.WEIXIN_REMOTE_STATIC_URL,
            # 是否调试模式
            "DEBUG": json.dumps(settings.DEBUG),
            # 当前时间
            "NOW": datetime.datetime.now(),
            # 前后端联合开发的静态资源路径, 这个变量可选配置
            "BK_STATIC_URL": getattr(settings, "BK_STATIC_URL", ""),
            # 是否使用blueking_language切换国际化语言
            "IS_DISPLAY_LANGUAGE_CHANGE": settings.IS_DISPLAY_LANGUAGE_CHANGE,
        }
    except Exception:  # pylint: disable=broad-except
        logger.exception(u"自定义模板上下文异常")
        raise
    return context



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸 (Blueking) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at https://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""
import json

from rest_framework import serializers
from rest_framework.settings import api_settings

from blueapps.core.exceptions import ParamValidationError


def format_serializer_errors(  # pylint: disable=too-many-locals,too-many-branches
    errors, fields, params, prefix="", return_all_errors=True
):
    """
    格式化序列化器的错误，对前端显示更为友好
    :param errors: serializer_errors
    :param fields: 校验的字段
    :param params: 参数
    :param prefix: 错误消息前缀
    :param return_all_errors: 是否返回所有错误消息
    :return:
    """
    message = {}
    cn_message = {}
    for key, field_errors in list(errors.items()):  # pylint: disable=too-many-nested-blocks
        sub_message = []
        try:
            label = fields[key].label
        except (KeyError, AttributeError):
            label = key

        if key == api_settings.NON_FIELD_ERRORS_KEY:
            sub_message.append(";".join(field_errors))
        elif key not in fields:
            sub_message.append(json.dumps(field_errors))
        else:
            field = fields[key]
            if (
                hasattr(field, "child")
                and isinstance(field_errors, list)
                and len(field_errors) > 0
                and not isinstance(field_errors[0], str)
            ):
                for index, sub_errors in enumerate(field_errors):
                    if sub_errors:
                        sub_format = format_serializer_errors(sub_errors, field.child.fields, params, prefix=prefix)
                        if not return_all_errors:
                            return f"{label}: {sub_format}"
                        temp_message = f"{prefix}第{index + 1}项:"
                        sub_message.append(
                            temp_message + sub_format[0] if isinstance(sub_format, tuple) else sub_format
                        )
            else:
                if isinstance(field_errors, dict):
                    if hasattr(field, "child"):
                        sub_format = format_serializer_errors(field_errors, field.child.fields, params, prefix=prefix)
                    else:
                        sub_format = format_serializer_errors(field_errors, field.fields, params, prefix=prefix)
                    if not return_all_errors:
                        return f"{label}: {sub_format}"
                    sub_message.append(sub_format)
                elif isinstance(field_errors, list):
                    for index, error in enumerate(field_errors):
                        error = error.format(**{key: params.get(key, "")})
                        sub_message.append("{index}.{error}".format(index=index + 1, error=error))
                        if not return_all_errors:
                            sub_message = ",".join(sub_message)
                            return f"{label}: {sub_message}"
        cn_message[str(label)] = ",".join(sub_message)
        message[key] = sub_message
    return json.dumps(message), json.dumps(cn_message)


def custom_params_valid(serializer, params, instance=None, many=False, partial=False):
    """序列化器自定义数据校验"""
    _serializer = serializer(data=params, many=many, instance=instance, partial=partial)
    try:
        _serializer.is_valid(raise_exception=True)
    except serializers.ValidationError:
        msg_tuple = format_serializer_errors(_serializer.errors, _serializer.fields, params)
        raise ParamValidationError(msg_tuple)
    if many:
        return list(_serializer.validated_data)

    return dict(_serializer.validated_data)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2022 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import functools
import json

from django.core.cache import cache
from django.core.serializers.json import DjangoJSONEncoder

from blueapps.utils.base import md5_sum
from blueapps.utils.constants import TimeEnum
from blueapps.utils.logger import logger


def using_cache(key: str, duration, need_md5=False):
    """
    缓存方法返回值
    :param key: key 名可以使用format进行格式
    :param duration: 过期时间
    :param need_md5: 缓冲是redis的时候 key不能带有空格等字符，需要用md5 hash一下
    :return: function
    """

    def decorator(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):

            try:
                actual_key = key.format(*args, **kwargs)
            except (IndexError, KeyError):
                actual_key = key

            logger.info(f"[using cache] build key => [{actual_key}] duration => [{duration}]")

            if need_md5:
                actual_key = md5_sum(actual_key)

            cache_result = cache.get(actual_key)

            if cache_result:
                return json.loads(cache_result)

            result = func(*args, **kwargs)
            if result:
                cache.set(actual_key, json.dumps(result, cls=DjangoJSONEncoder), duration)
            return result

        return inner

    return decorator


cache_half_minute = functools.partial(using_cache, duration=0.5 * TimeEnum.ONE_MINUTE_SECOND)
cache_one_minute = functools.partial(using_cache, duration=TimeEnum.ONE_MINUTE_SECOND)
cache_five_minute = functools.partial(using_cache, duration=5 * TimeEnum.ONE_MINUTE_SECOND)
cache_ten_minute = functools.partial(using_cache, duration=10 * TimeEnum.ONE_MINUTE_SECOND)
cache_one_hour = functools.partial(using_cache, duration=TimeEnum.ONE_HOUR_SECOND)
cache_one_day = functools.partial(using_cache, duration=TimeEnum.ONE_DAY_SECOND)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from django.db import models
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import ValidationError as DrfValidationError
from rest_framework.fields import empty
from rest_framework.serializers import ModelSerializer
from rest_framework.utils import model_meta

from blueapps.contrib.drf.serializer import format_serializer_errors
from blueapps.contrib.drf.utils.serializer_fields import CustomDateTimeField
from blueapps.core.exceptions import ParamValidationError
from blueapps.utils.db import MultiStrSplitCharField, MultiStrSplitTextField

"""
DRF 常用序列化器
"""


class GeneralSerializer(ModelSerializer):
    """
    自定义ModelSerializer序列化器
    """

    def __init__(self, instance=None, data=empty, **kwargs):
        self.serializer_field_mapping[models.DateTimeField] = CustomDateTimeField
        self.serializer_field_mapping[MultiStrSplitCharField] = serializers.ListField
        self.serializer_field_mapping[MultiStrSplitTextField] = serializers.ListField
        super(GeneralSerializer, self).__init__(instance=instance, data=data, **kwargs)

    def is_valid(self, raise_exception=False):
        """
        参数校验 返回为处理后的异常信息
        """
        try:
            super(GeneralSerializer, self).is_valid(raise_exception)
        # 捕获原生的参数校验异常，抛出SaaS封装的参数校验异常
        except DrfValidationError:
            if self._errors and raise_exception:
                raise ParamValidationError(format_serializer_errors(self.errors, self.fields, self.initial_data),)

        return not bool(self._errors)

    def create(self, validated_data):
        """
        序列化器创建数据
        """
        model_class = self.Meta.model

        info = model_meta.get_field_info(model_class)
        many_to_many = {}
        for field_name, relation_info in list(info.relations.items()):
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = validated_data.pop(field_name)

        try:
            instance = model_class.objects.create(**validated_data)
        except TypeError as exc:
            msg = (
                "Got a `TypeError` when calling `%s.objects.create()`. "
                "This may be because you have a writable field on the "
                "serializer class that is not a valid argument to "
                "`%s.objects.create()`. You may need to make the field "
                "read-only, or override the %s.create() method to handle "
                "this correctly.\nOriginal exception text was: %s."
                % (model_class.__name__, model_class.__name__, self.__class__.__name__, exc,)
            )
            raise TypeError(msg)

        # Save many-to-many relationships after the instance is created.
        if many_to_many:
            for field_name, value in list(many_to_many.items()):
                setattr(instance, field_name, value)

        return instance

    class Meta:
        """
        元数据定义
        """

        model = None


class PageSerializer(serializers.Serializer):
    """
    分页序列化器
    """

    page = serializers.IntegerField(help_text=_("页数"))
    pagesize = serializers.IntegerField(help_text=_("每页数量"))


class OrderingSerializer(serializers.Serializer):
    """
    排序序列化器
    """

    ordering = serializers.CharField(help_text=_("排序字段，`-`表示逆序，多个排序字段以`,`分隔"), required=False)


class SearchSerializer(serializers.Serializer):
    """
    关键字序列化器
    """

    search = serializers.CharField(help_text=_("模糊查找，多个查找关键字以`,`分隔"), required=False)


class PostFilterSerializer(serializers.Serializer):
    """
    ORM condition查询序列化器
    example:
    {
        "conditions": [
            {
                "fields": ["name"],
                "operator": "in",
                "values": ["user1", "user2"]
            }
        ]
    }
    """

    class ConditionSerializer(serializers.Serializer):
        """
        单个查询条件序列化器
        """

        fields = serializers.ListField(help_text=_("查询目标字段列表"), child=serializers.CharField(), min_length=1)
        operator = serializers.CharField(help_text=_("查询类型，可选：`in` `contains` `range` `gt(lt)` `gte(lte)..."))
        values = serializers.ListField(help_text=_("查询目标值列表"), min_length=1)

    conditions = serializers.ListField(help_text=_("查询条件"), child=ConditionSerializer(), required=False, default=[])



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import hashlib
import traceback
from collections import namedtuple
from contextlib import contextmanager

from blueapps.utils.logger import logger


@contextmanager
def ignored(*exceptions, **kwargs):
    """
    在忽略某异常下执行
    example:
        with ignored(Exception):
            # 不会抛出异常
            print(1/0)
    """
    try:
        yield
    except exceptions:
        if kwargs.get("log_exception", True):
            logger.warning(traceback.format_exc())


def dict_to_namedtuple(dic):
    """
    从dict转换到namedtuple
    """
    return namedtuple("AttrStore", list(dic.keys()))(**dic)


def choices_to_namedtuple(choices):
    """
    从django-model的choices转换到namedtuple
    """
    return dict_to_namedtuple(dict(choices))


def tuple_choices(tupl):
    """
    从django-model的choices转换到namedtuple
    """
    return [(t, t) for t in tupl]


def md5_sum(src_str: str):
    """
    计算md5_sum值
    @param src_str {str} 源字符串
    """
    md5 = hashlib.md5()
    md5.update(src_str.encode("utf-8"))
    md5_key = md5.hexdigest()
    return md5_key



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import random
import time

from django.conf import settings
from django.contrib import auth
from django.utils.deprecation import MiddlewareMixin

from blueapps.account.components.weixin.forms import WeixinAuthenticationForm
from blueapps.account.conf import ConfFixture
from blueapps.account.handlers.response import ResponseHandler

logger = logging.getLogger("component")


class WeixinLoginRequiredMiddleware(MiddlewareMixin):
    def process_view(self, request, view, args, kwargs):
        """
        可通过登录认证的方式，仅有两种
        1. 带有 login_exemp 标识的 view 函数
        2. 用户已成功 auth.login
        """
        # 框架前置中间件，已将识别的客户端信息填充进 request
        if not request.is_wechat():
            return None

        logger.debug("当前请求客户端为微信端")
        login_exempt = getattr(view, "login_exempt", False)

        if login_exempt or request.user.is_authenticated:
            return None

        user = self.authenticate(request)
        if user:
            return None

        self.set_state(request)
        handler = ResponseHandler(ConfFixture, settings)
        return handler.build_weixin_401_response(request)

    def process_response(self, request, response):
        return response

    def authenticate(self, request):
        form = WeixinAuthenticationForm(request.GET)

        if not form.is_valid():
            logger.error(
                u"微信请求链接，未检测到微信验证码，url：{}，params：{}".format(
                    request.path_info, request.GET
                )
            )
            return None

        code = form.cleaned_data["code"]
        state = form.cleaned_data["state"]
        logger.debug(u"微信请求链接，检测到微信验证码，code：{}，state：{}".format(code, state))

        if not self.valid_state(request, state):
            return None

        user = auth.authenticate(request=request, code=code, is_wechat=True)
        if user and user.username != request.user.username:
            auth.login(request, user)
        if request.user.is_authenticated:
            # 登录成功，确认登陆正常后退出
            return request.user
        return user

    def set_state(self, request, length=32):
        """
        生成随机数 state，表示客户端的当前状态，根据 oauth2.0 标准，在请求授权码时需要
        附带上的参数，认证服务器的回应必须一模一样包含这个参数，此处将 state 设置在
        session
        """
        allowed_chars = (
            "abcdefghijkmnpqrstuvwxyz" "ABCDEFGHIJKLMNPQRSTUVWXYZ" "0123456789"
        )
        state = "".join(random.choice(allowed_chars) for _ in range(length))
        request.session["WEIXIN_OAUTH_STATE"] = state
        request.session["WEIXIN_OAUTH_STATE_TIMESTAMP"] = time.time()
        return True

    def valid_state(self, request, state, expires_in=60):
        """
        验证微信认证服务器返回的 code & state 是否合法
        """
        raw_state = request.session.get("WEIXIN_OAUTH_STATE")
        raw_timestamp = request.session.get("WEIXIN_OAUTH_STATE_TIMESTAMP")

        if not raw_state or raw_state != state:
            logger.warning(
                u"验证 WEIXIN 服务器返回信息，state 不一致，"
                u"WEIXIN_OAUTH_STATE=%s，state=%s" % (raw_state, state),
            )
            return False

        if not raw_timestamp or time.time() - raw_timestamp > expires_in:
            logger.warning(
                u"验证 WEIXIN 服务器返回信息，state 过期，"
                u"WEIXIN_OAUTH_STATE_TIMESTAMP=%s" % raw_timestamp
            )
            return False

        request.session["WEIXIN_OAUTH_STATE"] = None
        request.session["WEIXIN_OAUTH_STATE_TIMESTAMP"] = None
        return True



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import sys
import uuid

from django.dispatch import Signal
from django.utils.deprecation import MiddlewareMixin

from blueapps.conf import settings
from blueapps.core.exceptions import AccessForbidden, ServerBlueException
from blueapps.settings import blueapps_settings
from blueapps.utils.local import inject_request_id, release_request_local, request_local

# since each thread has its own greenlet we can just use those as identifiers
# for the context.  If greenlets are not available we fall back to the
# current thread ident depending on where it is.
try:
    from greenlet import getcurrent as get_ident
except ImportError:
    try:
        from thread import get_ident
    except ImportError:
        from _thread import get_ident


class AccessorSignal(Signal):
    allowed_receiver = "blueapps.utils.request_provider.RequestProvider"

    def __init__(self, providing_args=None):
        Signal.__init__(self, providing_args)

    def connect(self, receiver, sender=None, weak=True, dispatch_uid=None):
        receiver_name = ".".join([receiver.__class__.__module__, receiver.__class__.__name__])
        if receiver_name != self.allowed_receiver:
            raise AccessForbidden(u"%s is not allowed to connect" % receiver_name)
        Signal.connect(self, receiver, sender, weak, dispatch_uid)


request_accessor = AccessorSignal()


class RequestProvider(MiddlewareMixin):
    """
    @summary: request事件接收者
    """

    _instance = None

    def __new__(cls, get_response):
        if cls._instance is None:
            cls._instance = super(RequestProvider, cls).__new__(cls)
        return cls._instance

    def __init__(self, get_response):
        super(RequestProvider, self).__init__(get_response)
        self._request_pool = {}
        request_accessor.connect(self)

    def process_request(self, request):
        request.is_mobile = lambda: bool(settings.RE_MOBILE.search(request.META.get("HTTP_USER_AGENT", "")))

        # 是否为合法的RIO请求
        request.is_rio = lambda: bool(
            request.META.get("HTTP_STAFFNAME", "")
            and getattr(settings, "RIO_TOKEN", None)
            and settings.RE_WECHAT.search(request.META.get("HTTP_USER_AGENT", ""))
        )

        # 是否为合法 WEIXIN 请求，必须符合两个条件，wx 客户端 & WX PAAS 域名
        request_origin_url = "{}://{}".format(request.scheme, request.get_host())
        request.is_wechat = lambda: (
            bool(settings.RE_WECHAT.search(request.META.get("HTTP_USER_AGENT", "")))
            and request_origin_url == settings.WEIXIN_BK_URL
            and not request.is_rio()
        )

        # JWT请求
        request.is_bk_jwt = lambda: bool(request.META.get("HTTP_X_BKAPI_JWT", ""))

        self._request_pool[get_ident()] = request
        request_local.request = request
        inject_request_id(request_local, request)
        return None

    def process_response(self, request, response):
        assert request is self._request_pool.pop(get_ident())
        release_request_local()
        return response

    def __call__(self, *args, **kwargs):
        """
        1）接受 signal 请求响应，
        2）继承 MiddlewareMixin.__call__ 兼容 djagno 1.10 之前中间件
        """
        from_signal = kwargs.get("from_signal", False)
        if from_signal:
            return self.get_request(**kwargs)
        else:
            return super(RequestProvider, self).__call__(args[0])

    def get_request(self, **kwargs):
        sender = kwargs.get("sender")
        if sender is None:
            sender = get_ident()
        if sender not in self._request_pool:
            raise ServerBlueException(u"get_request can't be called in a new thread.")
        return self._request_pool[sender]


def get_request():
    return request_accessor.send(get_ident(), from_signal=True)[0][1]


def get_x_request_id():
    x_request_id = ""
    http_request = get_request()
    if hasattr(http_request, "META"):
        meta = http_request.META
        x_request_id = meta.get("HTTP_X_REQUEST_ID", "") if isinstance(meta, dict) else ""
    return x_request_id


def get_local_request():
    return getattr(request_local, "request", None)


def get_local_request_id():
    return getattr(request_local, "request_id", str(uuid.uuid4()))


def get_or_create_local_request_id():
    if not hasattr(request_local, "request_id"):
        request_local.request_id = str(uuid.uuid4())
    return request_local.request_id


def get_request_username():
    operator = None
    try:
        operator = get_local_request().user.username
    except (IndexError, AttributeError):
        if blueapps_settings.NON_REQUEST_USERNAME_PROVIDER:
            operator = blueapps_settings.NON_REQUEST_USERNAME_PROVIDER()

    return operator



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from urllib.parse import urlencode

from django.http import HttpResponseRedirect, JsonResponse
from django.utils.translation import ugettext_lazy as _

from blueapps.account.utils.http import build_redirect_url
from blueapps.core.exceptions import BkJwtVerifyError, RioVerifyError
from blueapps.settings import blueapps_settings
from blueapps.utils.tools import resolve_login_url

try:
    from django.urls import reverse
except Exception:  # pylint: disable=broad-except
    from django.core.urlresolvers import reverse


class ResponseHandler(object):
    def __init__(self, _confFixture, _settings):
        """
        @param {object} confFixture Account Package Settings
        @param {object} settings Django User Settings
        """
        self._conf = _confFixture
        self._settings = _settings

    def build_403_response(self, msg):
        return HttpResponseRedirect(f'{reverse("account:login_forbidden")}?{urlencode({"hint_msg": msg})}')

    def build_401_response(self, request):

        # 强制要求进行跳转的方式
        if blueapps_settings.IS_AJAX_PLAIN_MODE and request.is_ajax():
            return self._build_ajax_401_response(request)

        # Just redirect to PAAS-LOGIN-PLATRORM no matter whether request.is_ajax
        if self._conf.HAS_PLAIN:
            if request.is_ajax():
                return self._build_ajax_401_response(request)
            else:
                return self._build_page_401_response(request)
        else:
            if request.is_ajax():
                context = {"has_plain": False}
                return JsonResponse(context, status=401)
            else:
                return self._build_page_401_response_platform(request)

    def _build_ajax_401_response(self, request):
        """
        Return 401 info, inlclude login_url to PAAS-LOGIN-PLATFORM,
        width & height for adjusting iframe window, login_url as
        http://xxx/login/?c_url=http%3A//xxx/t/data/&app_id=data
        """
        if blueapps_settings.AJAX_401_RESPONSE_FUNC:
            return blueapps_settings.AJAX_401_RESPONSE_FUNC(request)
        _next = request.build_absolute_uri(reverse("account:login_success"))

        if self._conf.ADD_CROSS_PREFIX:
            _next = self._conf.CROSS_PREFIX + _next

        _login_url = build_redirect_url(
            _next,
            resolve_login_url(self._conf.LOGIN_PLAIN_URL, request),
            self._conf.C_URL,
            extra_args=self._build_extra_args(),
        )

        context = {
            "login_url": _login_url,
            "width": self._conf.IFRAME_WIDTH,
            "height": self._conf.IFRAME_HEIGHT,
            "has_plain": True,
        }

        return JsonResponse(context, status=401)

    def _build_page_401_response(self, request):
        """
        Redirect to login page in self app, redirect url format as
        http://xxx:8000/account/login_page/?refer_url=http%3A//xxx%3A8000/
        """
        if blueapps_settings.PAGE_401_RESPONSE_FUNC:
            return blueapps_settings.PAGE_401_RESPONSE_FUNC(request)
        _next = request.build_absolute_uri()
        if request.method == "GET" and blueapps_settings.SPECIFIC_REDIRECT_KEY:
            _next = request.GET.get(blueapps_settings.SPECIFIC_REDIRECT_KEY) or _next
        _redirect = build_redirect_url(
            _next,
            resolve_login_url(self._conf.LOGIN_PLAIN_URL, request),
            self._conf.C_URL,
            extra_args=self._build_extra_args(),
        )
        return HttpResponseRedirect(_redirect)

    def _build_page_401_response_platform(self, request):
        """
        Directly redirect to PAAS-LOGIN-PLATFORM
        """
        if blueapps_settings.PAGE_401_RESPONSE_PLATFORM_FUNC:
            return blueapps_settings.PAGE_401_RESPONSE_PLATFORM_FUNC(request)
        _next = request.build_absolute_uri()
        if request.method == "GET" and blueapps_settings.SPECIFIC_REDIRECT_KEY:
            _next = request.GET.get(blueapps_settings.SPECIFIC_REDIRECT_KEY) or _next
        if self._conf.ADD_CROSS_PREFIX:
            _next = self._conf.CROSS_PREFIX + _next

        _login_url = build_redirect_url(
            _next,
            resolve_login_url(self._conf.LOGIN_URL, request),
            self._conf.C_URL,
            extra_args=self._build_extra_args(),
        )
        return HttpResponseRedirect(_login_url)

    def _build_extra_args(self):
        extra_args = None
        if self._conf.ADD_APP_CODE:
            extra_args = {self._conf.APP_KEY: getattr(self._settings, self._conf.SETTINGS_APP_KEY)}
        return extra_args

    def build_weixin_401_response(self, request):
        """
        todo，说明 url 格式
        """
        _login_url = self._conf.WEIXIN_OAUTH_URL
        _next = request.build_absolute_uri()
        if request.method == "GET" and blueapps_settings.SPECIFIC_REDIRECT_KEY:
            _next = request.GET.get(blueapps_settings.SPECIFIC_REDIRECT_KEY) or _next

        extra_args = {
            "appid": self._conf.WEIXIN_APP_ID,
            "response_type": "code",
            "scope": "snsapi_base",
            "state": request.session["WEIXIN_OAUTH_STATE"],
        }
        _redirect = build_redirect_url(_next, _login_url, "redirect_uri", extra_args=extra_args)
        return HttpResponseRedirect(_redirect)

    def build_rio_401_response(self, request):
        context = {
            "result": False,
            "code": RioVerifyError.ERROR_CODE,
            "message": _("您的登陆请求无法经智能网关正常检测，请与管理人员联系"),
        }
        return JsonResponse(context, status=401)

    def build_bk_jwt_401_response(self, request):
        """
        BK_JWT鉴权异常
        """
        context = {
            "result": False,
            "code": BkJwtVerifyError.ERROR_CODE,
            "message": _("您的登陆请求无法经BK JWT检测，请与管理人员联系"),
        }
        return JsonResponse(context, status=401)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2022 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import functools
import time
from contextlib import contextmanager

from django.conf import settings
from django.core.cache import cache

from blueapps.core.exceptions.base import LockError
from blueapps.utils.unique import uniqid


class BaseLock:
    def __init__(self, name, ttl=None):
        self.name = name
        # 默认60秒过期
        self.ttl = ttl or 60

    def acquire(self, _wait=None):
        raise NotImplementedError

    def release(self):
        raise NotImplementedError

    def __exit__(self, t, v, tb):
        self.release()

    def __enter__(self):
        self.acquire()
        return self


class RedisLock(BaseLock):
    __token = None

    def __init__(self, name, ttl=None):
        super(RedisLock, self).__init__(name, ttl)
        self.client = cache

    def acquire(self, _wait=0.001):
        token = uniqid()
        wait_until = time.time() + _wait
        while not self.client.set(self.name, token, timeout=self.ttl, nx=True):
            if time.time() < wait_until:
                time.sleep(0.01)
            else:
                return False

        self.__token = token
        return True

    def release(self):
        if not self.__token:
            return False
        token = self.client.get(self.name)
        if not token or token != self.__token:
            return False
        return self.client.delete(self.name)


@contextmanager
def service_lock(key_instance, **kwargs):
    lock = None
    lock_key = key_instance.get_key(**kwargs)
    try:
        lock = RedisLock(lock_key, key_instance.ttl)
        if lock.acquire(0.1):
            yield lock
        else:
            raise LockError(msg="{} is already locked".format(lock_key))
    except LockError as err:
        raise err

    finally:
        if lock is not None:
            lock.release()


def share_lock(ttl=600, identify=None):
    """
    装饰定时任务时需要放在periodic_task下面
    @periodic_task(run_every=crontab(minute="*/1"), queue="sync")
    # 不填参数需要带括号执行
    @share_lock()
    def demo():
        pass
    :param ttl:
    :param identify:
    :return:
    """

    def wrapper(func):
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            if not settings.USE_REDIS:
                return func(*args, **kwargs)
            token = str(time.time())
            # 防止函数重名导致方法失效，增加一个ID参数，可以通过ID参数屏蔽多模块函数名重复的问题
            # 例如，可以为`${module}_${method_used_for}`
            cache_key = "celery_%s" % func.__name__ if identify is None else identify
            client = cache
            lock_success = client.set(cache_key, token, timeout=ttl, nx=True)
            if not lock_success:
                return

            try:
                return func(*args, **kwargs)
            finally:
                if client.get(cache_key) == token:
                    client.delete(cache_key)

        return _inner

    return wrapper



import datetime
from django.conf import settings
import json
import os

# 网关相关默认配置项
APIGW_DEFAULT_SETTINGS = {
    # 发布时间，与发布版本拼接成，确保每一次发布都可以重新发布网关资源
    "DEPLOY_DATETIME": datetime.datetime.now().strftime("%Y%m%d%H%M%S"),
    # 发布版本 默认为1.0.0
    "RELEASE_VERSION": "1.0.0",
    # 网关所有者
    "MAINTAINERS": ["admin"],
    # 应用申请指定网关所有资源的权限 默认不需要
    "BK_APIGW_APPLY_API_NAME": [],
    # apigw_definition 模板默认输出位置
    "GENERATE_APIGW_DEFINITION_TARGET_DIR": "",
    # apigw_definition_file 模板文件名
    "APIGW_DEFINITION_FILE_NAME": "apigw-definition.yml",
    # api_resources 模板默认输出位置
    "GENERATE_API_RESOURCES_TARGET_DIR": "",
    # api_resources 模板文件名
    "API_RESOURCES_FILE_NAME": "api-resources.yml",
    # SaaS生产环境后端地址
    "BK_APIGW_API_PROD_HOST": os.getenv("BK_APIGW_API_PROD_HOST") or json.loads(
        os.getenv("BKPAAS_DEFAULT_PREALLOCATED_URLS", "{}")).get(
        "prod"),
    # 应用在SaaS部署后默认的预发布环境
    "BK_APIGW_API_STAG_HOST": os.getenv("BK_APIGW_API_STAG_HOST") or json.loads(
        os.getenv("BKPAAS_DEFAULT_PREALLOCATED_URLS", "{}")).get(
        "stag"),
}


class APIGWSettings:
    def __init__(self):
        # 默认网关配置项
        self.apigw_settings = APIGW_DEFAULT_SETTINGS
        # 对配置项进行补充
        self.get_delay_settings()
        # 获取用户定义的网关配置项
        self.user_settings = getattr(settings, "APIGW_DEFINITION_SETTINGS", {})
        # 合并配置项,以用户配置为主
        self.apigw_settings.update(self.user_settings)
        # apigw-manager解析配置文件需要使用settings，进行配置回写
        setattr(settings, "APIGW_DEFINITION_SETTINGS", self.apigw_settings)

    def __getattr__(self, key):
        if key not in self.apigw_settings:
            raise AttributeError
        value = self.apigw_settings[key]
        return value

    def get_delay_settings(self):
        """需要settings加载完毕后进行获取的配置项"""
        # 网关名称 默认为app code 不能有下划线
        self.apigw_settings["BK_APIGW_NAME"] = settings.APP_CODE
        # 网关主动授权给应用，添加访问网关所有资源的权限，默认授权当前saas
        self.apigw_settings["BK_APIGW_GRANT_APPS"] = [settings.APP_CODE, ]


apigw_settings = APIGWSettings()



import os
from django.core.management.base import BaseCommand
from django.core.management import call_command
from django.conf import settings
from blueapps.apigw_sync.apigw_sync_settings import apigw_settings


class Command(BaseCommand):
    help = 'Sync SaaS API Gateway'

    def add_arguments(self, parser):
        parser.add_argument('--definition_file', type=str, help='The relative path of the definition file')
        parser.add_argument('--resources_file', type=str, help='The relative path of the resources file')

    def handle(self, *args, **options):
        # 获取APP_CODE
        app_code = settings.APP_CODE
        # 获取文件路径
        definition_file = options['definition_file'] if options['definition_file'] \
            else os.path.join(apigw_settings.GENERATE_APIGW_DEFINITION_TARGET_DIR,
                              apigw_settings.APIGW_DEFINITION_FILE_NAME)
        resources_file = options['resources_file'] if options['resources_file'] \
            else os.path.join(apigw_settings.GENERATE_API_RESOURCES_TARGET_DIR,
                              apigw_settings.API_RESOURCES_FILE_NAME)
        # 获取文件绝对路径
        definition_file_path = os.path.join(settings.BASE_DIR, definition_file)
        resources_file_path = os.path.join(settings.BASE_DIR, resources_file)
        api_name = apigw_settings.BK_APIGW_NAME

        # 同步网关基本信息
        call_command("sync_apigw_config", file=definition_file_path, api_name=api_name)
        self.stdout.write(
            self.style.SUCCESS("[%s] Successfully called sync_apigw_config with definition: %s" % (
                app_code, definition_file_path)))

        # 同步网关环境信息
        call_command("sync_apigw_stage", file=definition_file_path, api_name=api_name)
        self.stdout.write(
            self.style.SUCCESS(
                "[%s] Successfully called sync_apigw_stage with definition: %s" % (app_code, definition_file_path)))

        # 同步网关资源 --del
        call_command("sync_apigw_resources", "--delete", file=resources_file_path, api_name=api_name)
        self.stdout.write(
            self.style.SUCCESS("[%s] Successfully called sync_apigw_resources with resources: %s" % (
                app_code, resources_file_path)))

        # 创建资源版本并发布
        call_command("create_version_and_release_apigw", "--generate-sdks", file=definition_file_path,
                     api_name=api_name, stage=[settings.ENVIRONMENT])
        self.stdout.write(self.style.SUCCESS(
            "[%s] Successfully called create_version_and_release_apigw with definition: %s" % (
                app_code, definition_file_path)))

        # 为应用主动授权 可选
        call_command("grant_apigw_permissions", file=definition_file_path, api_name=api_name)
        self.stdout.write(self.style.SUCCESS(
            "[%s] Successfully called grant_apigw_permissions with definition: %s" % (
                app_code, definition_file_path)))

        # 获取网关公钥
        call_command("fetch_apigw_public_key", api_name=api_name)
        self.stdout.write(self.style.SUCCESS("[%s] Successfully called fetch_apigw_public_key" % app_code))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
from collections import OrderedDict

from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response


class CustomPageNumberPagination(PageNumberPagination):
    """
    自定义分页
    """

    # 默认一页拉取的数量
    page_size = 100

    # 获取URL参数中传入的页码key字段
    page_query_param = "page"

    # 指定单页最大值的字段
    page_size_query_param = "page_size"

    # 设置单次拉取的最大值
    max_page_size = 1000

    def get_paginated_response(self, data):
        """
        返回分页后的Response
        """
        return Response(
            OrderedDict(
                [
                    ("page", self.page.number),
                    ("num_pages", self.page.paginator.num_pages),
                    ("total", self.page.paginator.count),
                    ("results", data),
                ]
            )
        )

    def get_paginated_data(self, data):
        """
        生成分页数据
        """
        return OrderedDict(
            [
                ("page", self.page.number),
                ("num_pages", self.page.paginator.num_pages),
                ("total", self.page.paginator.count),
                ("results", data),
            ]
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""
import typing

from django.conf import settings

try:
    from bkcrypto.contrib.django.fields import SymmetricTextField as CustomTextField


except ImportError:
    # 没有 bkcrypto-python-sdk 相关依赖，此时不启用加密
    from django.db import models

    class CustomTextField(models.TextField):
        def __init__(self, *args, using: typing.Optional[str] = None, **kwargs):
            self.using = using or "default"

            super(CustomTextField, self).__init__(*args, **kwargs)


class SymmetricTextField(CustomTextField):
    def get_decrypted_value(self, value):
        if value is None:
            return super().get_decrypted_value(value)
        return super().get_decrypted_value(str(value))

    @property
    def enable_encryption(self) -> bool:
        """
        开发框架是否已指定 cipher 实例对 SymmetricTextField 所涉及的 DB 字段进行加密
        不指定时采取明文落库策略
        """
        return bool(settings.BLUEAPPS_ENABLE_DB_ENCRYPTION)

    def encrypt(self, plaintext: str) -> str:
        if self.enable_encryption:
            return super().encrypt(plaintext)
        return plaintext

    def decrypt(self, ciphertext_with_prefix: str) -> str:
        if self.enable_encryption:
            return super().decrypt(ciphertext_with_prefix)
        return ciphertext_with_prefix



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import os
import random
import string

from blueapps.conf.default_settings import APP_CODE, BASE_DIR
from blueapps.patch.log import get_paas_v2_logging_config_dict

APP_CODE = os.environ.get("APP_ID", APP_CODE)


def set_log_level(settings_module):
    log_level = settings_module.get("LOG_LEVEL", "INFO")
    is_local = settings_module.get("IS_LOCAL", False)
    is_open_saas_v2 = settings_module.get("is_open_saas_v2", lambda: False)

    if is_open_saas_v2():
        bk_log_dir = settings_module.get("BK_LOG_DIR", "/data/apps/logs/")
        logging = get_paas_v2_logging_config_dict(is_local, bk_log_dir, log_level)
    else:
        logging = get_logging_config_dict(settings_module)

    return logging


def get_logging_config_dict(settings_module):
    log_class = "logging.handlers.RotatingFileHandler"
    log_level = settings_module.get("LOG_LEVEL", "INFO")

    if settings_module.get("IS_LOCAL", False):
        log_dir = os.path.join(os.path.dirname(BASE_DIR), "logs", APP_CODE)
        log_name_prefix = os.getenv("BKPAAS_LOG_NAME_PREFIX", APP_CODE)
        logging_format = {
            "format": (
                "%(levelname)s [%(asctime)s] %(pathname)s "
                "%(lineno)d %(funcName)s %(process)d %(thread)d "
                "\n \t %(message)s \n"
            ),
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    else:
        log_dir = settings_module.get("LOG_DIR_PREFIX", "/app/v3logs/")
        rand_str = "".join(random.sample(string.ascii_letters + string.digits, 4))
        log_name_prefix = "{}-{}".format(os.getenv("BKPAAS_PROCESS_TYPE"), rand_str)

        logging_format = {
            "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
            "fmt": (
                "%(levelname)s %(asctime)s %(pathname)s %(lineno)d " "%(funcName)s %(process)d %(thread)d %(message)s"
            ),
        }
    os.makedirs(log_dir, exist_ok=True)

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {"verbose": logging_format, "simple": {"format": "%(levelname)s %(message)s"}},
        "handlers": {
            "null": {"level": "DEBUG", "class": "logging.NullHandler"},
            "console": {"level": "DEBUG", "class": "logging.StreamHandler", "formatter": "simple"},
            "root": {
                "class": log_class,
                "formatter": "verbose",
                "filename": os.path.join(log_dir, "%s-django.log" % log_name_prefix),
                "maxBytes": 1024 * 1024 * 100,
                "backupCount": 5,
            },
            "component": {
                "class": log_class,
                "formatter": "verbose",
                "filename": os.path.join(log_dir, "%s-component.log" % log_name_prefix),
                "maxBytes": 1024 * 1024 * 100,
                "backupCount": 5,
            },
            "mysql": {
                "class": log_class,
                "formatter": "verbose",
                "filename": os.path.join(log_dir, "%s-mysql.log" % log_name_prefix),
                "maxBytes": 1024 * 1024 * 100,
                "backupCount": 5,
            },
            "celery": {
                "class": log_class,
                "formatter": "verbose",
                "filename": os.path.join(log_dir, "%s-celery.log" % log_name_prefix),
                "maxBytes": 1024 * 1024 * 100,
                "backupCount": 5,
            },
            "blueapps": {
                "class": log_class,
                "formatter": "verbose",
                "filename": os.path.join(log_dir, "%s-django.log" % log_name_prefix),
                # TODO blueapps log 等待平台提供单独的路径
                # log_dir, '%s-blueapps.log' % log_name_prefix),
                "maxBytes": 1024 * 1024 * 100,
                "backupCount": 5,
            },
        },
        "loggers": {
            "django": {"handlers": ["null"], "level": "INFO", "propagate": True},
            "django.server": {"handlers": ["console"], "level": log_level, "propagate": True},
            "django.request": {"handlers": ["root"], "level": "ERROR", "propagate": True},
            "django.db.backends": {"handlers": ["mysql"], "level": log_level, "propagate": True},
            # the root logger ,用于整个project的logger
            "root": {"handlers": ["root"], "level": log_level, "propagate": True},
            # 组件调用日志
            "component": {"handlers": ["component"], "level": log_level, "propagate": True},
            "celery": {"handlers": ["celery"], "level": log_level, "propagate": True},
            # other loggers...
            # blueapps
            "blueapps": {"handlers": ["blueapps"], "level": log_level, "propagate": True},
            # 普通app日志
            "app": {"handlers": ["root"], "level": log_level, "propagate": True},
        },
    }



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging

from django.conf import settings
from django.contrib import auth
from django.utils.deprecation import MiddlewareMixin

from blueapps.account.components.rio.forms import RioAuthenticationForm
from blueapps.account.conf import ConfFixture
from blueapps.account.handlers.response import ResponseHandler

logger = logging.getLogger("component")


class RioLoginRequiredMiddleware(MiddlewareMixin):
    def process_view(self, request, view, args, kwargs):
        """
        可通过登录认证的请求：
        1. 带有 RIO HTTP HEADER
        2. RIO签名正确
        """
        # 框架前置中间件，已将识别的客户端信息填充进 request
        if not hasattr(request, "is_rio") or not request.is_rio():
            return None

        logger.debug("当前请求是否经过RIO转发")
        login_exempt = getattr(view, "login_exempt", False)

        if login_exempt or request.user.is_authenticated:
            return None

        # 每次请求都需要做校验
        user = self.authenticate(request)
        if user:
            return None

        handler = ResponseHandler(ConfFixture, settings)
        return handler.build_rio_401_response(request)

    def process_response(self, request, response):
        return response

    def authenticate(self, request):
        form = RioAuthenticationForm(request.META)
        if not form.is_valid():
            logger.debug(
                u"[RIO]未检测到请求标识，url：{}，params：{}, errors: {}".format(
                    request.path_info, request.META, form.errors
                )
            )
            return None

        timestamp = form.cleaned_data["HTTP_TIMESTAMP"]
        signature = form.cleaned_data["HTTP_SIGNATURE"]
        staff_id = form.cleaned_data["HTTP_STAFFID"]
        staff_name = form.cleaned_data["HTTP_STAFFNAME"]
        x_ext_data = form.cleaned_data["HTTP_X_EXT_DATA"]
        x_rio_seq = form.cleaned_data["HTTP_X_RIO_SEQ"]
        logger.debug(u"RIO请求链接，检测到验证码：%s" % (signature))

        user = auth.authenticate(
            request=request,
            timestamp=timestamp,
            signature=signature,
            staff_id=staff_id,
            staff_name=staff_name,
            x_ext_data=x_ext_data,
            x_rio_seq=x_rio_seq,
        )
        if user and user.username != request.user.username:
            auth.login(request, user)
        if request.user.is_authenticated:
            # 登录成功，确认登陆正常后退出
            return request.user
        return user



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from __future__ import absolute_import

import importlib
import os

from blueapps.conf import get_settings_from_module, get_config_module

config = get_config_module()

locals().update(get_settings_from_module(config))

BASE_DIR = locals()["BASE_DIR"]
RUN_VER = locals()["RUN_VER"]
if "BK_URL" not in locals().keys():
    BK_URL = None

# 根据平台加载对应变量
try:
    site_mod = importlib.import_module("blueapps.conf.sites.%s" % RUN_VER)
except ImportError:
    raise ImportError(u"unknown RUN_VER: %s" % RUN_VER)
for _setting in dir(site_mod):
    if _setting.isupper():
        locals()[_setting] = getattr(site_mod, _setting)

# Inherit from paas environment
# all = [
#     'BKPAAS_LOG_NAME_PREFIX',
#     'BKPAAS_ENGINE_APP_NAME',
#     'BKPAAS_ENGINE_REGION',
#     'BKPAAS_ENVIRONMENT',
#     'BKPAAS_URL',
#     'BKPAAS_SUB_PATH',
#     'BKPAAS_REMOTE_STATIC_URL',
#     'BKPAAS_WEIXIN_URL',
#     'BKPAAS_WEIXIN_REMOTE_STATIC_URL',
# ]

# 蓝鲸平台URL
BK_URL = os.getenv("BKPAAS_URL", BK_URL)

# 蓝鲸开发者页面
BK_DEV_URL = BK_URL

# 站点URL
SITE_URL = os.getenv("BKPAAS_SUB_PATH", "/")

# 远程静态文件URL
REMOTE_STATIC_URL = os.getenv("BKPAAS_REMOTE_STATIC_URL", "%s/static_api/" % BK_URL)

# 判断是否为本地开发环境
IS_LOCAL = not os.getenv("BKPAAS_ENVIRONMENT", False)

# static root and dirs to find blueapps static
if not IS_LOCAL:
    STATIC_ROOT = "staticfiles"
    FORCE_SCRIPT_NAME = SITE_URL

    # 开启子域名时静态文件统一使用子域名访问，同时可以由用户通过环境变量修改入口URL
    app_subdomains = os.getenv("BKAPP_ENGINE_APP_DEFAULT_SUBDOMAINS", None)
    if not app_subdomains:
        app_subdomains = os.getenv("BKPAAS_ENGINE_APP_DEFAULT_SUBDOMAINS", None)
    # 存在该变量，而且不是空字符串
    if app_subdomains is not None and app_subdomains != "":
        STATIC_URL = "//%s/static/" % app_subdomains.split(";")[0]
    else:
        STATIC_URL = "%sstatic/" % FORCE_SCRIPT_NAME
else:
    STATIC_ROOT = os.path.join(BASE_DIR, "staticfiles")
    STATIC_URL = "/static/"

# About Log
LOG_NAME_PREFIX = os.getenv("BKPAAS_LOG_NAME_PREFIX")

# About whitenoise
WHITENOISE_STATIC_PREFIX = "/static/"

# Rabbitmq & Celery
if "RABBITMQ_VHOST" in os.environ:
    RABBITMQ_VHOST = os.getenv("RABBITMQ_VHOST")
    RABBITMQ_PORT = os.getenv("RABBITMQ_PORT")
    RABBITMQ_HOST = os.getenv("RABBITMQ_HOST")
    RABBITMQ_USER = os.getenv("RABBITMQ_USER")
    RABBITMQ_PASSWORD = os.getenv("RABBITMQ_PASSWORD")
    BROKER_URL = "amqp://{user}:{password}@{host}:{port}/{vhost}".format(
        user=RABBITMQ_USER,
        password=RABBITMQ_PASSWORD,
        host=RABBITMQ_HOST,
        port=RABBITMQ_PORT,
        vhost=RABBITMQ_VHOST,
    )

# WEIXIN Settings

# 微信 URL 前缀
WEIXIN_URL_PREFIX = "weixin"

# APP 微信 ROOT URL
WEIXIN_SITE_URL = "{}{}/".format(SITE_URL, WEIXIN_URL_PREFIX)

# 平台微信 URL 域名
WEIXIN_BK_URL = os.getenv("BKPAAS_WEIXIN_URL", "https://mt.bk.tencent.com")

# APP 微信本地静态资源目录
# TODO 环境变量中无WEXIN_STATIC_URL或BKPAAS_WEIXIN_STATIC_URL
WEIXIN_STATIC_URL = os.getenv("BKPAAS_WEIXIN_STATIC_URL", "%sstatic/weixin/" % SITE_URL)

# APP 微信远程静态资源目录
WEIXIN_REMOTE_STATIC_URL = os.getenv(
    "BKPAAS_WEIXIN_REMOTE_STATIC_URL", "%s/static_api/" % WEIXIN_BK_URL
)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback

from blueapps.core.exceptions.base import MethodError
from django.conf import settings
from django.contrib.auth.backends import ModelBackend
from django.db import IntegrityError

from blueapps import metrics
from blueapps.account import get_user_model
from blueapps.account.conf import ConfFixture
from blueapps.account.utils.http import send
from blueapps.utils import client
from blueapps.utils.tools import resolve_login_url

logger = logging.getLogger("component")

ROLE_TYPE_ADMIN = "1"

TOKEN_TYPE = "bk_token"


class TokenBackend(ModelBackend):
    def authenticate(self, request=None, bk_token=None):
        logger.debug("Enter in TokenBackend")
        # 判断是否传入验证所需的bk_token,没传入则返回None
        if not bk_token:
            return None

        verify_result, username = self.verify_bk_token(bk_token, request)
        # 判断bk_token是否验证通过,不通过则返回None
        if not verify_result:
            return None

        user_model = get_user_model()
        try:
            user, _ = user_model.objects.get_or_create(username=username)
            get_user_info_result, user_info = self.get_user_info(bk_token)
            # 判断是否获取到用户信息,获取不到则返回None
            if not get_user_info_result:
                return None
            user.set_property(key="qq", value=user_info.get("qq", ""))
            user.set_property(key="language", value=user_info.get("language", ""))
            user.set_property(key="time_zone", value=user_info.get("time_zone", ""))
            user.set_property(key="role", value=user_info.get("role", ""))
            user.set_property(key="phone", value=user_info.get("phone", ""))
            user.set_property(key="email", value=user_info.get("email", ""))
            user.set_property(key="wx_userid", value=user_info.get("wx_userid", ""))
            user.set_property(key="chname", value=user_info.get("chname", ""))

            # 用户如果不是管理员，则需要判断是否存在平台权限，如果有则需要加上
            if not user.is_superuser and not user.is_staff:
                role = user_info.get("role", "")
                is_admin = True if str(role) == ROLE_TYPE_ADMIN else False
                user.is_superuser = is_admin
                user.is_staff = is_admin
                user.save()

            return user

        except IntegrityError:
            logger.exception(traceback.format_exc())
            logger.exception("get_or_create UserModel fail or update_or_create UserProperty")
            return None
        except Exception:  # pylint: disable=broad-except
            logger.exception(traceback.format_exc())
            logger.exception("Auto create & update UserModel fail")
            return None

    @staticmethod
    def get_user_info(bk_token):
        """
        请求平台ESB接口获取用户信息
        @param bk_token: bk_token
        @type bk_token: str
        @return:True, {
            u'message': u'\u7528\u6237\u4fe1\u606f\u83b7\u53d6\u6210\u529f',
            u'code': 0,
            u'data': {
                u'qq': u'',
                u'wx_userid': u'',
                u'language': u'zh-cn',
                u'username': u'test',
                u'time_zone': u'Asia/Shanghai',
                u'role': 2,
                u'phone': u'11111111111',
                u'email': u'test',
                u'chname': u'test'
            },
            u'result': True,
            u'request_id': u'eac0fee52ba24a47a335fd3fef75c099'
        }
        @rtype: bool,dict
        """
        api_params = {"bk_token": bk_token}

        try:
            response = client.bk_login.get_user(api_params)
        except Exception as err:  # pylint: disable=broad-except
            logger.exception("Abnormal error in get_user_info...:%s" % err)
            return False, {}

        if response.get("result") is True:
            # 由于v1,v2的get_user存在差异,在这里屏蔽字段的差异,返回字段相同的字典
            origin_user_info = response.get("data", "")
            user_info = dict()
            # v1,v2字段相同的部分
            user_info["wx_userid"] = origin_user_info.get("wx_userid", "")
            user_info["language"] = origin_user_info.get("language", "")
            user_info["time_zone"] = origin_user_info.get("time_zone", "")
            user_info["phone"] = origin_user_info.get("phone", "")
            user_info["chname"] = origin_user_info.get("chname", "")
            user_info["email"] = origin_user_info.get("email", "")
            user_info["qq"] = origin_user_info.get("qq", "")
            # v2版本特有的字段
            if settings.DEFAULT_BK_API_VER == "v2":
                user_info["username"] = origin_user_info.get("bk_username", "")
                user_info["role"] = origin_user_info.get("bk_role", "")
            # v1版本特有的字段
            elif settings.DEFAULT_BK_API_VER == "":
                user_info["username"] = origin_user_info.get("username", "")
                user_info["role"] = origin_user_info.get("role", "")
            return True, user_info
        else:
            error_msg = response.get("message", "")
            error_data = response.get("data", "")
            logger.error("Failed to Get User Info: error=%(err)s, ret=%(ret)s" % {"err": error_msg, "ret": error_data})
            return False, {}

    def verify_bk_token(self, bk_token, request=None):
        """
        请求VERIFY_URL,认证bk_token是否正确
        @param bk_token: "_FrcQiMNevOD05f8AY0tCynWmubZbWz86HslzmOqnhk"
        @type bk_token: str
        @return: False,None True,username
        @rtype: bool,None/str
        """
        try:
            return self.verify_bk_token_through_esb(bk_token)
        except (NotImplementedError, MethodError, AttributeError):
            # 对应版本 esb 不支持 client.bk_login.is_login 接口的情况
            return self.verify_bk_token_through_verify_url(bk_token, request)

    @staticmethod
    def verify_bk_token_through_esb(bk_token):
        api_params = {"bk_token": bk_token}

        try:
            with metrics.observe(
                metrics.BLUEAPPS_USER_TOKEN_VERIFY_DURATION, hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE,
            ):
                response = client.bk_login.is_login(api_params)
        except (NotImplementedError, MethodError, AttributeError):
            raise
        except Exception:  # pylint: disable=broad-except
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="unknow_execption_raise",
            ).inc()
            logger.exception("Abnormal error in verify_bk_token...")
            return False, None

        if response.get("result"):
            data = response.get("data")
            username = data.get("bk_username")
            return True, username
        else:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="verify_fail"
            ).inc()
            error_msg = response.get("message", "")
            error_data = response.get("data", "")
            logger.error("Fail to verify bk_token, error={}, ret={}".format(error_msg, error_data))
            return False, None

    @staticmethod
    def verify_bk_token_through_verify_url(bk_token, request=None):
        api_params = {"bk_token": bk_token}

        try:
            with metrics.observe(
                metrics.BLUEAPPS_USER_TOKEN_VERIFY_DURATION, hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE
            ):
                response = send(
                    resolve_login_url(ConfFixture.VERIFY_URL, request, "http"), "GET", api_params, verify=False,
                )
        except Exception:  # pylint: disable=broad-except
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="unknow_execption_raise"
            ).inc()
            logger.exception("Abnormal error in verify_bk_token...")
            return False, None

        if response.get("result"):
            data = response.get("data")
            username = data.get("username")
            return True, username
        else:
            metrics.BLUEAPPS_USER_TOKEN_VERIFY_FAILED_TOTAL.labels(
                hostname=metrics.HOSTNAME, token_type=TOKEN_TYPE, err="verify_fail"
            ).inc()
            error_msg = response.get("message", "")
            error_data = response.get("data", "")
            logger.error("Fail to verify bk_token, error={}, ret={}".format(error_msg, error_data))
            return False, None
